# Copyright (c) 2020-2021 Micro Focus or one of its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function, division, absolute_import

from ..message import BackendMessage


class EndOfBatchResponse(BackendMessage):
    message_id = b'J'

    def __init__(self, data):
        BackendMessage.__init__(self)


BackendMessage.register(EndOfBatchResponse)



# Copyright (c) 2018-2021 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Vertica Frontend/Backend Protocol

All communication between frontend (client) and backend (server) is through a
stream of messages.

Message Format: The first byte of a message identifies the message type, and
the next four bytes specify the length of the rest of the message (this length
count includes itself, but not the message-type byte). The remaining contents
of the message are determined by the message type. An exception is the Startup
message sent by frontend, which has no message-type byte.

All data in Vertica is represented as UTF-8, so the frontend has to convert data
going to the backend from Python text string into UTF-8, and to convert data
coming from the backend from UTF-8 into Python text string.
"""

from __future__ import print_function, division, absolute_import

from abc import ABCMeta
from struct import pack

from ..messages import *


class Message(object):
    __metaclass__ = ABCMeta

    def __init__(self):
        pass

    @property
    def message_id(self):
        raise NotImplementedError("no default message_id")

    def _bytes_to_message(self, msg):

        if isinstance(msg, list):
            msg = ''.join(msg)

        if hasattr(msg, 'bytesize'):
            bytesize = msg.bytesize + 4
        else:
            bytesize = len(msg) + 4

        message_size = pack('!I', bytesize)
        if self.message_id is not None:
            msg_with_size = self.message_id + message_size + msg
        else:
            msg_with_size = message_size + msg

        return msg_with_size

    def __str__(self):
        return self.__class__.__name__


# noinspection PyAbstractClass
class BackendMessage(Message):
    __metaclass__ = ABCMeta
    _message_id_map = {}

    @classmethod
    def from_type(cls, type_, data):
        klass = cls._message_id_map.get(type_)
        if klass is not None:
            return klass(data)
        else:
            from .backend_messages import Unknown
            return Unknown(type_, data)

    @staticmethod
    def register(cls):
        # TODO replace _message_id() with that
        assert issubclass(cls, BackendMessage), \
            "{0} is not subclass of BackendMessage".format(cls.__name__)
        assert cls.message_id not in BackendMessage._message_id_map, \
            "can't write the same key twice: {0}".format(cls.message_id)

        BackendMessage._message_id_map[cls.message_id] = cls


# noinspection PyAbstractClass
class FrontendMessage(Message):
    __metaclass__ = ABCMeta

    def fetch_message(self):
        """Generator for getting the message's content"""
        raise NotImplementedError("fetch_bytes has no default implementation")


# noinspection PyAbstractClass
class BulkFrontendMessage(FrontendMessage):
    __metaclass__ = ABCMeta

    def read_bytes(self):
        return b''

    def get_message(self):
        bytes_ = self.read_bytes()
        return self._bytes_to_message(bytes_)

    def fetch_message(self):
        yield self.get_message()


# noinspection PyAbstractClass
class StreamFrontendMessage(FrontendMessage):
    __metaclass__ = ABCMeta

    def stream_bytes(self):
        raise NotImplementedError("stream_bytes has no default implementation")

    def stream_message(self):
        for bytes_ in self.stream_bytes():
            yield self._bytes_to_message(bytes_)

    def fetch_message(self):
        for message in self.stream_message():
            yield message



# Copyright (c) 2018-2021 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from __future__ import print_function, division, absolute_import

import re

from aiovertica.mixins.notice_response_attr import _NoticeResponseAttrMixin


#############################################
# dbapi errors
#############################################
class Error(Exception):
    pass


# noinspection PyShadowingBuiltins
class Warning(Exception):
    pass


class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass


class InternalError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class DataError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    pass


#
# Other Errors
#


class TimedOutError(OperationalError):
    pass


class ConnectionError(DatabaseError):
    pass

class KerberosError(ConnectionError):
    pass

class SSLNotSupported(ConnectionError):
    pass


class MessageError(InternalError):
    pass


class EmptyQueryError(ProgrammingError):
    pass


class QueryError(_NoticeResponseAttrMixin, ProgrammingError):
    """
    An error occurred associated with a query had been issued.

    This error is perhaps the most commonly encountered error type, associated
    with failures during query execution, invalid SQL statements, and more.

    Much of the details of the error are available as properties, including the
    SQL statement that had been issued (`sql`).
    """


    def __init__(self, error_response, sql):
        self.error_response = error_response
        self.sql = sql
        ProgrammingError.__init__(self,
                                  "{0}, SQL: {1}".format(error_response.error_message(),
                                                         repr(self.one_line_sql())))

    def one_line_sql(self):
        if self.sql:
            return re.sub(r"[\r\n]+", ' ', self.sql)
        else:
            return ''

    def __reduce__(self):
        # Workaround to make these exception instances (un)picklable. This must
        # be redefined by any subclass that changes the signature of __init__
        #
        # This workaround applies to python3. https://bugs.python.org/issue37489
        return (
            type(self),
            (self.error_response, self.sql),
            self.__dict__,
        )

    @property
    def _notice_attrs(self):
        # provided for _NoticeResponseAttrMixin
        if self.error_response is None:
            return {}
        return self.error_response._notice_attrs or {}

    @classmethod
    def from_error_response(cls, error_response, sql):
        klass = QUERY_ERROR_CLASSES.get(error_response.sqlstate, None)
        if klass is None:
            klass = cls
        return klass(error_response, sql)


class LockFailure(QueryError):
    pass


class InsufficientResources(QueryError):
    pass


class OutOfMemory(QueryError):
    pass


class VerticaSyntaxError(QueryError):
    pass


class MissingSchema(QueryError):
    pass


class MissingRelation(QueryError):
    pass


class MissingColumn(QueryError):
    pass


class CopyRejected(QueryError):
    pass


class PermissionDenied(QueryError):
    pass


class InvalidDatetimeFormat(QueryError):
    pass


class DuplicateObject(QueryError):
    pass


class QueryCanceled(QueryError):
    pass


class ConnectionFailure(QueryError):
    pass


class LostConnectivityFailure(QueryError):
    pass


QUERY_ERROR_CLASSES = {
    '55V03': LockFailure,
    '53000': InsufficientResources,
    '53200': OutOfMemory,
    '42601': VerticaSyntaxError,
    '3F000': MissingSchema,
    '42V01': MissingRelation,
    '42703': MissingColumn,
    '22V04': CopyRejected,
    '42501': PermissionDenied,
    '22007': InvalidDatetimeFormat,
    '42710': DuplicateObject,
    '57014': QueryCanceled,
    '08006': ConnectionFailure,
    'V1001': LostConnectivityFailure
}



# Copyright (c) 2018-2021 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Bind message

In the extended query protocol, the frontend sends a Bind message to bind values
to parameter placeholders present in an existing prepared statement.

The response is either BindComplete or ErrorResponse.
"""

from __future__ import print_function, division, absolute_import

from struct import pack
from six import string_types

from ..message import BulkFrontendMessage
from aiovertica.datatypes import VerticaType
from aiovertica.compat import as_bytes

BACKSLASH = b'\\'
BACKSLASH_ESCAPE = b'\\134'


class Bind(BulkFrontendMessage):
    message_id = b'B'

    def __init__(self, portal_name, prepared_statement_name, parameter_values, parameter_type_oids):
        BulkFrontendMessage.__init__(self)
        self._portal_name = portal_name
        self._prepared_statement_name = prepared_statement_name
        self._parameter_values = parameter_values
        self._parameter_type_oids = parameter_type_oids

    def read_bytes(self):
        utf_portal_name = self._portal_name.encode('utf-8')
        utf_prepared_statement_name = self._prepared_statement_name.encode('utf-8')

        bytes_ = pack('!{0}sx{1}sx'.format(len(utf_portal_name), len(utf_prepared_statement_name)),
                      utf_portal_name, utf_prepared_statement_name)

        # Parameter format codes -- use the default format (text)
        bytes_ += pack('!H', 0)

        # Number of parameters
        bytes_ += pack('!H', len(self._parameter_type_oids))

        param_bytes_ = b''
        for oid, val in zip(self._parameter_type_oids, self._parameter_values):
            # Parameter type oids
            bytes_ += pack('!I', oid)
            # Parameter values
            if val is None:  # -1 indicates a NULL parameter value
                param_bytes_ += pack('!i', -1)
            elif oid in (VerticaType.BINARY, VerticaType.VARBINARY, VerticaType.LONGVARBINARY):
                # Encode binary data as UTF8 bytes
                val = as_bytes(val)
                # Escape the byte value \ with "\134"(octal for backslash)
                val = val.replace(BACKSLASH, BACKSLASH_ESCAPE)
                param_bytes_ += pack('!I{0}s'.format(len(val)), len(val), val)
            else:
                # Convert input to string
                if oid == VerticaType.BOOL:
                    val = '1' if str(val).lower() in ('t', 'true', 'y', 'yes', '1') else '0'
                elif not isinstance(val, (string_types, bytes)):
                    val = str(val)
                # Encode string as UTF8 bytes
                val = val.encode('utf-8') if not isinstance(val, bytes) else val
                param_bytes_ += pack('!I{0}s'.format(len(val)), len(val), val)

        bytes_ += param_bytes_

        # Result column format codes -- use the default format (text)
        bytes_ += pack('!H', 0)

        return bytes_



# Copyright (c) 2018-2021 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, division, absolute_import

import os
import hashlib
from struct import pack

from ..message import BulkFrontendMessage
from ..backend_messages.authentication import Authentication
from aiovertica.compat import as_bytes

if os.name == 'nt':
    from . import crypt_windows as crypt
else:
    import crypt


class Password(BulkFrontendMessage):
    message_id = b'p'

    def __init__(self, password, auth_method=None, options=None):
        BulkFrontendMessage.__init__(self)

        self._password = as_bytes(password)
        self._options = options or {}
        if auth_method is not None:
            self._auth_method = auth_method
        else:
            self._auth_method = Authentication.CLEARTEXT_PASSWORD

    def encoded_password(self):

        if self._auth_method == Authentication.CLEARTEXT_PASSWORD:
            return self._password
        elif self._auth_method == Authentication.CRYPT_PASSWORD:
            return crypt.crypt(self._password, self._options['salt'])
        elif self._auth_method in (Authentication.MD5_PASSWORD,
                                   Authentication.HASH,
                                   Authentication.HASH_MD5,
                                   Authentication.HASH_SHA512):
            # Encodes user/password/salt information in the following way:
            #   MD5(MD5(password + user) + salt)
            #   SHA512(SHA512(password + userSalt) + salt)
            useMD5 = self._auth_method in (Authentication.MD5_PASSWORD, Authentication.HASH_MD5)
            user = self._options['user'].encode('utf-8') if useMD5 else self._options['usersalt']
            for key in (user, self._options['salt']):
                m = hashlib.md5() if useMD5 else hashlib.sha512()
                m.update(self._password + key)
                hexdigest = m.hexdigest()
                self._password = hexdigest.encode('utf-8')
            prefix = b'md5' if useMD5 else b'sha512'
            return prefix + self._password
        elif self._auth_method == Authentication.GSS:
            return self._password
        else:
            raise ValueError("unsupported authentication method: {0}".format(self._auth_method))

    def read_bytes(self):
        encoded_pw = self.encoded_password()
        # Vertica server handles GSS messages differently from other passwords
        if self._auth_method == Authentication.GSS:
            bytes_ = pack('{0}s'.format(len(encoded_pw)), encoded_pw)
        else:
            bytes_ = pack('{0}sx'.format(len(encoded_pw)), encoded_pw)
        return bytes_



# Copyright (c) 2018-2021 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Functions for Python 2 vs. 3 compatibility.
## Conversion routines
In addition to the functions below, `as_str` converts an object to a `str`.
@@as_bytes
@@as_text
@@as_str_any
## Types
The compatibility module also provides the following types:
* `bytes_or_text_types`
* `complex_types`
* `integral_types`
* `real_types`
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import Union

import six as _six


def as_bytes(bytes_or_text, encoding='utf-8'):
    """Converts either bytes or unicode to `bytes`, using utf-8 encoding for text.
    Args:
      bytes_or_text: A `bytes`, `bytearray`, `str`, or `unicode` object.
      encoding: A string indicating the charset for encoding unicode.
    Returns:
      A `bytes` object.
    Raises:
      TypeError: If `bytes_or_text` is not a binary or unicode string.
    """
    if isinstance(bytes_or_text, _six.text_type):
        return bytes_or_text.encode(encoding)
    elif isinstance(bytes_or_text, bytearray):
        return bytes(bytes_or_text)
    elif isinstance(bytes_or_text, bytes):
        return bytes_or_text
    else:
        raise TypeError('Expected binary or unicode string, got %r' %
                        (bytes_or_text,))


def as_text(bytes_or_text: Union[str, bytes], encoding: str = 'utf-8') -> str:
    """Returns the given argument as a unicode string.
    Args:
      bytes_or_text: A `bytes`, `bytearray`, `str`, or `unicode` object.
      encoding: A string indicating the charset for decoding unicode.
    Returns:
      A `unicode` (Python 2) or `str` (Python 3) object.
    Raises:
      TypeError: If `bytes_or_text` is not a binary or unicode string.
    """
    if isinstance(bytes_or_text, _six.text_type):
        return str(bytes_or_text)
    elif isinstance(bytes_or_text, (bytes, bytearray)):
        return bytes_or_text.decode(encoding)
    else:
        raise TypeError('Expected binary or unicode string, got %r' % bytes_or_text)


# Convert an object to a `str` in both Python 2 and 3.
if _six.PY2:
    as_str = as_bytes
else:
    as_str = as_text


def as_str_any(value):
    """Converts to `str` as `str(value)`, but use `as_str` for `bytes`.
    Args:
      value: A object that can be converted to `str`.
    Returns:
      A `str` object.
    """
    if isinstance(value, (bytes, bytearray)):
        return as_str(value)
    else:
        return str(value)


# Either bytes or text.
bytes_or_text_types = (bytes, bytearray, _six.text_type)

_allowed_symbols = [
    'as_str',
    'bytes_or_text_types',
]



# Copyright (c) 2018-2021 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from __future__ import print_function, division, absolute_import

import re
from collections import namedtuple
from datetime import date, datetime
from decimal import Decimal
from six import PY2
from uuid import UUID

# noinspection PyCompatibility,PyUnresolvedReferences
from dateutil import parser, tz

from aiovertica import errors
from aiovertica.datatypes import VerticaType, getDisplaySize, getPrecision, getScale
from aiovertica.compat import as_str, as_text


YEARS_RE = re.compile(r"^([0-9]+)-")


# these methods are bad...
#
# a few timestamp with tz examples:
# 2013-01-01 00:00:00
# 2013-01-01 00:00:00+00
# 2013-01-01 00:00:00.01+00
# 2013-01-01 00:00:00.00001+00
#
# Vertica stores all data in UTC:
#   "TIMESTAMP WITH TIMEZONE (TIMESTAMPTZ) data is stored in GMT (UTC) by
#    converting data from the current local time zone to GMT."
# Vertica fetches data in local timezone:
#   "When TIMESTAMPTZ data is used, data is converted back to use the current
#    local time zone"
# If vertica boxes are on UTC, you should never have a non +00 offset (as
# far as I can tell) ie. inserting '2013-01-01 00:00:00.01 EST' to a
# timestamptz type stores: 2013-01-01 05:00:00.01+00
#       select t AT TIMEZONE 'America/New_York' returns: 2012-12-31 19:00:00.01
def timestamp_parse(s):
    s = as_str(s)
    try:
        dt = _timestamp_parse(s)
    except ValueError:
        # Value error, year might be over 9999
        year_match = YEARS_RE.match(s)
        if year_match:
            year = year_match.groups()[0]
            dt = _timestamp_parse_without_year(s[len(year) + 1:])
            dt = dt.replace(year=min(int(year), 9999))
        else:
            raise errors.DataError('Timestamp value not supported: %s' % s)

    return dt


def _timestamp_parse(s):
    if len(s) == 19:
        return datetime.strptime(s, '%Y-%m-%d %H:%M:%S')
    return datetime.strptime(s, '%Y-%m-%d %H:%M:%S.%f')


def _timestamp_parse_without_year(s):
    if len(s) == 14:
        return datetime.strptime(s, '%m-%d %H:%M:%S')
    return datetime.strptime(s, '%m-%d %H:%M:%S.%f')


def timestamp_tz_parse(s):
    s = as_str(s)
    # if timezone is simply UTC...
    if s.endswith('+00'):
        # remove time zone
        ts = timestamp_parse(s[:-3].encode(encoding='utf-8', errors='strict'))
        ts = ts.replace(tzinfo=tz.tzutc())
        return ts
    # other wise do a real parse (slower)
    return parser.parse(s)


def date_parse(s):
    """
    Parses value of a DATE type.
    :param s: string to parse into date
    :return: an instance of datetime.date
    :raises NotSupportedError when a date Before Christ is encountered
    """
    s = as_str(s)
    if s.endswith(' BC'):
        raise errors.NotSupportedError('Dates Before Christ are not supported. Got: {0}'.format(s))

    # Value error, year might be over 9999
    return date(*map(lambda x: min(int(x), 9999), s.split('-')))


def time_parse(s):
    s = as_str(s)
    if len(s) == 8:
        return datetime.strptime(s, '%H:%M:%S').time()
    return datetime.strptime(s, '%H:%M:%S.%f').time()

def binary_data_parse(s):
    """
    Parses text value of a BINARY/VARBINARY/LONG VARBINARY type.
    :param s: bytearray
    :return: bytes
    """
    buf = []
    i = 0
    while i < len(s):
        c = s[i: i+1]
        if c == b'\\':
            c2 = s[i+1: i+2]
            if c2 == b'\\':  # escaped \
                if PY2: c = str(c)
                i += 2
            else:   # A \xxx octal string
                c = chr(int(str(s[i+1: i+4]), 8)) if PY2 else bytes([int(s[i+1: i+4], 8)])
                i += 4
        else:
            if PY2: c = str(c)
            i += 1
        buf.append(c)
    return b''.join(buf)


# Type casting of SQL types bytes representation into Python objects
def vertica_type_cast(type_code, unicode_error):
    typecaster = {
        VerticaType.UNKNOWN: bytes,
        VerticaType.BOOL: lambda s: s == b't',
        VerticaType.INT8: lambda s: int(s),
        VerticaType.FLOAT8: lambda s: float(s),
        VerticaType.CHAR: lambda s: s.decode('utf-8', unicode_error),
        VerticaType.VARCHAR: lambda s: s.decode('utf-8', unicode_error),
        VerticaType.DATE: date_parse,
        VerticaType.TIME: time_parse,
        VerticaType.TIMESTAMP: timestamp_parse,
        VerticaType.TIMESTAMPTZ: timestamp_tz_parse,
        VerticaType.INTERVAL: bytes,
        VerticaType.TIMETZ: bytes,
        VerticaType.NUMERIC: lambda s: Decimal(s.decode('utf-8', unicode_error)),
        VerticaType.VARBINARY: binary_data_parse,
        VerticaType.UUID: lambda s: UUID(s.decode('utf-8', unicode_error)),
        VerticaType.INTERVALYM: bytes,
        VerticaType.LONGVARCHAR: lambda s: s.decode('utf-8', unicode_error),
        VerticaType.LONGVARBINARY: binary_data_parse,
        VerticaType.BINARY: binary_data_parse
    }
    return typecaster.get(type_code, bytes)


ColumnTuple = namedtuple('Column', ['name', 'type_code', 'display_size', 'internal_size',
                                    'precision', 'scale', 'null_ok'])


class Column(object):
    def __init__(self, col, unicode_error='strict'):
        # Describe one query result column
        self.name = col['name']
        self.type_code = col['data_type_oid']
        self.type_name = col['data_type_name']
        self.display_size = getDisplaySize(col['data_type_oid'], col['type_modifier'])
        self.internal_size = col['data_type_size']
        self.precision = getPrecision(col['data_type_oid'], col['type_modifier'])
        self.scale = getScale(col['data_type_oid'], col['type_modifier'])
        self.null_ok = col['null_ok']
        self.is_identity = col['is_identity']
        self.converter = vertica_type_cast(self.type_code, unicode_error)
        self.props = ColumnTuple(self.name, self.type_code, self.display_size, self.internal_size,
                                 self.precision, self.scale, self.null_ok)

    def convert(self, s):
        if s is None:
            return
        return self.converter(s)

    def __str__(self):
        return as_str(str(self.props))

    def __unicode__(self):
        return as_text(str(self.props))

    def __repr__(self):
        return as_str(str(self.props))

    def __iter__(self):
        for prop in self.props:
            yield prop

    def __getitem__(self, key):
        return self.props[key]



# Copyright (c) 2018-2021 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import print_function, division, absolute_import

from ..message import BackendMessage


class CloseComplete(BackendMessage):
    message_id = b'3'

    def __init__(self, data):
        BackendMessage.__init__(self)


BackendMessage.register(CloseComplete)



# Copyright (c) 2020-2021 Micro Focus or one of its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import print_function, division, absolute_import

import errno
import os


def ensure_dir_exists(filepath):
    """Ensure that a directory exists

    If it doesn't exist, try to create it and protect against a race condition
    if another process is doing the same.
    """
    directory = os.path.dirname(filepath)
    if directory != '' and not os.path.exists(directory):
        try:
            os.makedirs(directory)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

def check_file_readable(filename):
    """Ensure this is a readable file"""
    if not os.path.exists(filename):
        raise OSError('{} does not exist'.format(filename))
    elif not os.path.isfile(filename):
        raise OSError('{} is not a file'.format(filename))
    elif not os.access(filename, os.R_OK):
        raise OSError('{} is not readable'.format(filename))

def check_file_writable(filename):
    """Ensure this is a writable file. If the file doesn't exist,
       ensure its directory is writable.
    """
    if os.path.exists(filename):
        if not os.path.isfile(filename):
            raise OSError('{} is not a file'.format(filename))
        if not os.access(filename, os.W_OK):
            raise OSError('{} is not writable'.format(filename))
    # If target does not exist, check permission on parent dir
    ensure_dir_exists(filename)
    pdir = os.path.dirname(filename)
    if not pdir:
        pdir = '.'
    if not os.path.isdir(pdir):
        raise OSError('{} is not a directory'.format(pdir))
    if not os.access(pdir, os.W_OK):
        raise OSError('Directory {} is not writable'.format(pdir))




# Copyright (c) 2018-2021 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
# Copyright (c) 2021 Ivan Galin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from __future__ import print_function, division, absolute_import

import asyncio
import datetime
import glob
import inspect
import logging
import re
import sys
import traceback
import types
from decimal import Decimal
from io import IOBase
from tempfile import NamedTemporaryFile, SpooledTemporaryFile, TemporaryFile
from uuid import UUID
from collections import OrderedDict
from typing import Any, AsyncGenerator, Callable, Dict, KeysView, List, Optional, \
    TYPE_CHECKING, \
    Tuple, Type, \
    Union

# _TemporaryFileWrapper is an undocumented implementation detail, so
# import defensively.
try:
    from tempfile import _TemporaryFileWrapper
except ImportError:
    _TemporaryFileWrapper = None

import six
# noinspection PyUnresolvedReferences,PyCompatibility
from six import binary_type, text_type, string_types, integer_types, BytesIO, StringIO
from six.moves import zip

from aiovertica import errors, messages, os_utils
from aiovertica.compat import as_text
from aiovertica.column import Column

if TYPE_CHECKING:
    from aiovertica.connection import Connection


# A note regarding support for temporary files:
#
# Since Python 2.6, the tempfile module offers three kinds of temporary
# files:
#
#   * NamedTemporaryFile
#   * SpooledTemporaryFile
#   * TemporaryFile
#
# NamedTemporaryFile is not a class, but a function that returns
# an instance of the tempfile._TemporaryFileWrapper class.
# _TemporaryFileWrapper is a direct subclass of object.
#
#   * https://github.com/python/cpython/blob/v3.8.0/Lib/tempfile.py#L546
#   * https://github.com/python/cpython/blob/v3.8.0/Lib/tempfile.py#L450
#
# SpooledTemporaryFile is a class that is a direct subclass of object.
#
#   * https://github.com/python/cpython/blob/v3.8.0/Lib/tempfile.py#L623
#
# TemporaryFile is a class that is either NamedTemporaryFile or an
# indirect subclass of io.IOBase, depending on the platform.
#
#   * https://bugs.python.org/issue33762
#   * https://github.com/python/cpython/blob/v3.8.0/Lib/tempfile.py#L552-L555
#   * https://github.com/python/cpython/blob/v3.8.0/Lib/tempfile.py#L606-L608
#   * https://github.com/python/cpython/blob/v3.8.0/Lib/tempfile.py#L617-L618
#
# As a result, for Python 2.6 and newer, it seems the best way to test
# for a file-like object inclusive of temporary files is via:
#
#   isinstance(obj, (IOBase, SpooledTemporaryFile, _TemporaryFileWrapper))

# Of the following "types", only include those that are classes in
# file_type so that isinstance(obj, file_type) won't fail. As of Python
# 3.8 only IOBase, SpooledTemporaryFile and _TemporaryFileWrapper are
# classes, but if future Python versions implement NamedTemporaryFile
# and TemporaryFile as classes, the following code should account for
# that accordingly.
file_type = tuple(
    type_ for type_ in [
        IOBase,
        NamedTemporaryFile,
        SpooledTemporaryFile,
        TemporaryFile,
        _TemporaryFileWrapper,
    ]
    if inspect.isclass(type_)
)
if six.PY2:
    # noinspection PyUnresolvedReferences
    file_type = file_type + (file,)


RE_NAME_BASE = u"[0-9a-zA-Z_][\\w\\d\\$_]*"
RE_NAME = u'(("{0}")|({0}))'.format(RE_NAME_BASE)
RE_BASIC_INSERT_STAT = (
    u"\\s*INSERT\\s+INTO\\s+(?P<target>({0}\\.)?{0})"
    u"\\s*\\(\\s*(?P<variables>{0}(\\s*,\\s*{0})*)\\s*\\)"
    u"\\s+VALUES\\s*\\(\\s*(?P<values>(.|\\s)*)\\s*\\)").format(RE_NAME)
END_OF_RESULT_RESPONSES = (messages.CommandComplete, messages.PortalSuspended)
END_OF_BATCH_RESPONSES = (messages.WriteFile, messages.EndOfBatchResponse)
DEFAULT_BUFFER_SIZE = 131072

CursorType = Optional[Union[Type[List], Type[Dict]]]
FileType = Union[
    IOBase,
    NamedTemporaryFile,
    SpooledTemporaryFile,
    TemporaryFile,
    _TemporaryFileWrapper
]


class Cursor:
    # NOTE: this is used in executemany and is here for pandas compatibility
    _insert_statement = re.compile(RE_BASIC_INSERT_STAT, re.U | re.I)

    def __init__(
        self,
        connection: 'Connection',
        logger: logging.Logger,
        cursor_type: Optional[CursorType] = None,
        # TODO: May be Enum type...
        unicode_error: Optional[str] = None,
    ):
        self.connection = connection
        self._logger = logger
        self.cursor_type = cursor_type
        self.unicode_error = unicode_error if unicode_error is not None else 'strict'
        self._closed = False
        self._message = None
        self.operation = None
        self.prepared_sql = None  # last statement been prepared
        self.prepared_name = "s0"
        self.error = None
        self._sql_literal_adapters = {}

        #
        # dbapi attributes
        #
        self.description = None
        self.rowcount = -1
        self.arraysize = 1

    #############################################
    # supporting `with` statements
    #############################################
    async def __aenter__(self) -> 'Cursor':
        return self

    async def __aexit__(
        self,
        type_: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[types.TracebackType]
    ) -> None:
        await self.close()

    #############################################
    # dbapi methods
    #############################################
    # noinspection PyMethodMayBeStatic
    def callproc(
        self,
        procname: str,
        parameters: Optional[Dict[str, Any]] = None
    ) -> None:
        raise errors.NotSupportedError('Cursor.callproc() is not implemented')

    async def close(self) -> None:
        self._logger.info('Close the cursor')
        if not self.closed() and self.prepared_sql:
            await self._close_prepared_statement()
        self._closed = True

    async def execute(
        self,
        operation: str,
        parameters: Optional[Union[List[Any], Tuple[Any]]] = None,
        use_prepared_statements: Optional[bool] = None,
        # TODO: Pass type
        copy_stdin=None,
        buffer_size: int = DEFAULT_BUFFER_SIZE,
    ) -> 'Cursor':
        if self.closed():
            raise errors.InterfaceError('Cursor is closed')

        await self.flush_to_query_ready()

        operation = as_text(operation)
        self.operation = operation

        self.rowcount = -1

        if copy_stdin is None:
            self.copy_stdin_list = []
        elif (isinstance(copy_stdin, list) and
              all(callable(getattr(i, 'read', None)) for i in copy_stdin)):
            self.copy_stdin_list = copy_stdin
        elif callable(getattr(copy_stdin, 'read', None)):
            self.copy_stdin_list = [copy_stdin]
        else:
            raise TypeError(
                "Cursor.execute 'copy_stdin' parameter should be"
                " a file-like object or a list of file-like objects"
            )

        self.buffer_size = buffer_size  # For copy-local read and write

        use_prepared = bool(
            self.connection.options['use_prepared_statements']
            if use_prepared_statements is None else use_prepared_statements
        )
        if use_prepared:
            # Execute the SQL as prepared statement (server-side bindings)
            if parameters and not isinstance(parameters, (list, tuple)):
                raise TypeError("Execute parameters should be a list/tuple")

            # If the SQL has not been prepared, prepare the SQL
            if operation != self.prepared_sql:
                await self._prepare(operation)
                self.prepared_sql = operation  # the prepared statement is kept

            # Bind the parameters and execute
            await self._execute_prepared_statement([parameters])
        else:
            # Execute the SQL directly (client-side bindings)
            if parameters:
                operation = self.format_operation_with_parameters(operation, parameters)
            await self._execute_simple_query(operation)

        return self

    async def executemany(
        self,
        operation: str,
        seq_of_parameters: Union[List[Any], Tuple[Any]],
        use_prepared_statements: Optional[bool] = None,
    ) -> None:
        if not isinstance(seq_of_parameters, (list, tuple)):
            raise TypeError("seq_of_parameters should be list/tuple")

        if self.closed():
            raise errors.InterfaceError('Cursor is closed')

        await self.flush_to_query_ready()

        operation = as_text(operation)
        self.operation = operation

        use_prepared = bool(
            self.connection.options['use_prepared_statements']
            if use_prepared_statements is None else use_prepared_statements
        )

        if use_prepared:
            # Execute the SQL as prepared statement (server-side bindings)
            if len(seq_of_parameters) == 0:
                raise ValueError("seq_of_parameters should not be empty")
            if not all(isinstance(elem, (list, tuple)) for elem in seq_of_parameters):
                raise TypeError("Each seq_of_parameters element should be a list/tuple")
            # If the SQL has not been prepared, prepare the SQL
            if operation != self.prepared_sql:
                await self._prepare(operation)
                self.prepared_sql = operation  # the prepared statement is kept

            # Bind the parameters and execute
            await self._execute_prepared_statement(seq_of_parameters)
        else:
            m = self._insert_statement.match(operation)
            if m:
                target = as_text(m.group('target'))

                variables = as_text(m.group('variables'))
                variables = ",".join([
                    variable.strip().strip('"')
                    for variable in variables.split(",")
                ])

                values = as_text(m.group('values'))
                values = ",".join([
                    value.strip().strip('"') for value in values.split(",")
                ])
                seq_of_values = [
                    self.format_operation_with_parameters(
                        values, parameters, is_copy_data=True
                    )
                    for parameters in seq_of_parameters
                ]
                data = "\n".join(seq_of_values)

                copy_statement = (
                    f"COPY {target} ({variables}) FROM STDIN DELIMITER ','"
                    f" ENCLOSED BY '\"' ENFORCELENGTH ABORT ON ERROR"
                    f"{' NO COMMIT' if not self.connection.autocommit else ''}"
                )
                await self.copy(copy_statement, data)

            else:
                raise NotImplementedError(
                    "executemany is implemented for simple INSERT statements only"
                )

    async def fetchone(self) -> Optional[Union[List[Any], OrderedDict[str, Any]]]:
        while True:
            if isinstance(self._message, messages.DataRow):
                if self.rowcount == -1:
                    self.rowcount = 1
                else:
                    self.rowcount += 1
                row = self.row_formatter(self._message)
                # fetch next message
                self._message = await self.connection.read_message()
                return row
            elif isinstance(self._message, messages.RowDescription):
                self.description = [
                    Column(fd, self.unicode_error)
                    for fd in self._message.fields
                ]
            elif isinstance(self._message, messages.ReadyForQuery):
                return None
            elif isinstance(self._message, END_OF_RESULT_RESPONSES):
                return None
            elif isinstance(self._message, messages.EmptyQueryResponse):
                pass
            elif isinstance(self._message, messages.VerifyFiles):
                await self._handle_copy_local_protocol()
            elif isinstance(self._message, messages.EndOfBatchResponse):
                pass
            elif isinstance(self._message, messages.CopyDoneResponse):
                pass
            elif isinstance(self._message, messages.ErrorResponse):
                raise errors.QueryError.from_error_response(
                    self._message, self.operation
                )
            else:
                raise errors.MessageError(
                    f'Unexpected fetchone() state: {type(self._message).__name__}'
                )

            self._message = self.connection.read_message()

    async def fetchmany(
        self,
        size: Optional[int] = None
    ) -> List[
        Optional[Union[List[Any], OrderedDict[str, Any]]]
    ]:
        if not size:
            size = self.arraysize
        results = []
        while True:
            row = await self.fetchone()
            if not row:
                break
            results.append(row)
            if len(results) >= size:
                break
        return results

    async def fetchall(self) -> List[Optional[Union[List[Any], OrderedDict[str, Any]]]]:
        data = []
        async for item in self.iterate():
            data.append(item)
        return data

    async def nextset(self) -> bool:
        """
        Skip to the next available result set, discarding any remaining rows
        from the current result set.

        If there are no more result sets, this method returns False. Otherwise,
        it returns a True and subsequent calls to the fetch*() methods will
        return rows from the next result set.
        """
        # skip any data for this set if exists
        await self.flush_to_end_of_result()

        if self._message is None:
            return False
        elif isinstance(self._message, END_OF_RESULT_RESPONSES):
            # there might be another set, read next message to find out
            self._message = await self.connection.read_message()
            if isinstance(self._message, messages.RowDescription):
                self.description = [
                    Column(fd, self.unicode_error)
                    for fd in self._message.fields
                ]
                self._message = await self.connection.read_message()
                if isinstance(self._message, messages.VerifyFiles):
                    await self._handle_copy_local_protocol()
                self.rowcount = -1
                return True
            elif isinstance(self._message, messages.BindComplete):
                self._message = await self.connection.read_message()
                self.rowcount = -1
                return True
            elif isinstance(self._message, messages.ReadyForQuery):
                return False
            elif isinstance(self._message, END_OF_RESULT_RESPONSES):
                # result of a DDL/transaction
                self.rowcount = -1
                return True
            elif isinstance(self._message, messages.ErrorResponse):
                raise errors.QueryError.from_error_response(
                    self._message, self.operation
                )
            else:
                raise errors.MessageError(
                    f'Unexpected nextset() state after END_OF_RESULT_RESPONSES:'
                    f' {self._message}'
                )
        elif isinstance(self._message, messages.ReadyForQuery):
            # no more sets left to be read
            return False
        else:
            raise errors.MessageError(
                f'Unexpected nextset() state: {self._message}'
            )

    def setinputsizes(self, sizes: int) -> None:
        pass

    def setoutputsize(self, size: int, column: Optional[Any] = None) -> None:
        pass

    #############################################
    # non-dbapi methods
    #############################################
    def closed(self) -> bool:
        return self._closed or self.connection.closed()

    def cancel(self) -> None:
        # Cancel is a session-level operation, cursor-level API does not make
        # sense. Keep this API for backward compatibility.
        raise errors.NotSupportedError(
            'Cursor.cancel() is deprecated. Call Connection.cancel() '
            'to cancel the current database operation.'
        )

    async def iterate(
        self
    ) -> AsyncGenerator[
        None, Optional[Union[List[Any], OrderedDict[str, Any]]]
    ]:
        row = await self.fetchone()
        while row:
            yield row
            row = await self.fetchone()

    async def copy(
        self,
        sql: str,
        data: Union[binary_type, text_type, FileType],
        **kwargs: KeysView[Any]
    ) -> None:
        """

        EXAMPLE:
        >> with open("/tmp/file.csv", "rb") as fs:
        >>     cursor.copy("COPY table(field1,field2) FROM STDIN DELIMITER ',' ENCLOSED BY ''''",
        >>                 fs, buffer_size=65536)

        """
        sql = as_text(sql)

        if self.closed():
            raise errors.InterfaceError('Cursor is closed')

        await self.flush_to_query_ready()

        if isinstance(data, binary_type):
            stream = BytesIO(data)
        elif isinstance(data, text_type):
            stream = StringIO(data)
        elif isinstance(data, file_type) or callable(getattr(data, 'read', None)):
            stream = data
        else:
            raise TypeError(f"Not valid type of data {type(data)}")

        # TODO: check sql is a valid `COPY FROM STDIN` SQL statement

        self._logger.info(f'Execute COPY statement: [{sql}]')
        # Execute a `COPY FROM STDIN` SQL statement
        await self.connection.write(messages.Query(sql))

        buffer_size = (
            kwargs['buffer_size'] if 'buffer_size' in kwargs else DEFAULT_BUFFER_SIZE
        )

        while True:
            message = await self.connection.read_message()

            self._message = message
            if isinstance(message, messages.ErrorResponse):
                raise errors.QueryError.from_error_response(message, sql)
            elif isinstance(message, messages.ReadyForQuery):
                break
            elif isinstance(message, messages.CommandComplete):
                pass
            elif isinstance(message, messages.CopyInResponse):
                try:
                    await self._send_copy_data(stream, buffer_size)
                except Exception as e:
                    # COPY termination: report the cause of failure to the backend
                    await self.connection.write(messages.CopyFail(str(e)))
                    self._logger.error(str(e))
                    await self.flush_to_query_ready()
                    raise errors.DataError(
                        f'Failed to send a COPY data stream: {e}'
                    )

                # Successful termination for COPY
                await self.connection.write(messages.CopyDone())
            else:
                raise errors.MessageError(f'Unexpected message: {message}')

        if self.error is not None:
            raise self.error

    def object_to_sql_literal(self, py_obj: object) -> str:
        return self.object_to_string(py_obj, False)

    def register_sql_literal_adapter(
        self,
        obj_type: object,
        adapter_func: Callable[..., Any]
    ) -> None:
        if not callable(adapter_func):
            raise TypeError(
                "Cannot register this sql literal adapter. The adapter is not callable."
            )
        self._sql_literal_adapters[obj_type] = adapter_func

    #############################################
    # internal
    #############################################
    async def flush_to_query_ready(self) -> None:
        # if the last message isn't empty or ReadyForQuery, read all remaining messages
        if self._message is None or isinstance(self._message, messages.ReadyForQuery):
            return

        while True:
            message = await self.connection.read_message()
            if isinstance(message, messages.ReadyForQuery):
                self._message = message
                break
            elif isinstance(message, messages.VerifyFiles):
                self._message = message
                await self._handle_copy_local_protocol()

    async def flush_to_end_of_result(self) -> None:
        # if the last message isn't empty or END_OF_RESULT_RESPONSES,
        # read messages until it is
        if (
            self._message is None
            or isinstance(self._message, messages.ReadyForQuery)
            or isinstance(self._message, END_OF_RESULT_RESPONSES)
        ):
            return

        while True:
            message = await self.connection.read_message()
            if isinstance(message, END_OF_RESULT_RESPONSES):
                self._message = message
                break

    # TODO: Set type for row_data
    def row_formatter(self, row_data) -> Union[List[Any], OrderedDict[str, Any]]:
        if self.cursor_type is None:
            return self.format_row_as_array(row_data)
        elif self.cursor_type in (list, 'list'):
            return self.format_row_as_array(row_data)
        elif self.cursor_type in (dict, 'dict'):
            return self.format_row_as_dict(row_data)
        else:
            raise TypeError('Unrecognized cursor_type: {0}'.format(self.cursor_type))

    # TODO: Set type for row_data
    def format_row_as_dict(self, row_data) -> OrderedDict[str, Any]:
        return OrderedDict(
            (descr.name, descr.convert(value))
            for descr, value in zip(self.description, row_data.values)
        )

    # TODO: Set type for row_data
    def format_row_as_array(self, row_data) -> List[Any]:
        return [descr.convert(value)
                for descr, value in zip(self.description, row_data.values)]

    def object_to_string(
        self,
        py_obj: object,
        is_copy_data: bool
    ) -> str:
        """Return the SQL representation of the object as a string"""
        if type(py_obj) in self._sql_literal_adapters and not is_copy_data:
            adapter = self._sql_literal_adapters[type(py_obj)]
            result = adapter(py_obj)
            if not isinstance(result, (string_types, bytes)):
                raise TypeError(
                    f"Unexpected return type of {type(py_obj)} adapter:"
                    f" {type(result)}, expected a string type."
                )
            return as_text(result)

        if isinstance(py_obj, type(None)):
            return '' if is_copy_data else 'NULL'
        elif isinstance(py_obj, bool):
            return str(py_obj)
        elif isinstance(py_obj, (string_types, bytes)):
            return self.format_quote(as_text(py_obj), is_copy_data)
        elif isinstance(py_obj, (integer_types, float, Decimal)):
            return str(py_obj)
        elif isinstance(py_obj, tuple):  # tuple and namedtuple
            elements = [None] * len(py_obj)
            for i in range(len(py_obj)):
                elements[i] = self.object_to_string(py_obj[i], is_copy_data)
            return f"({','.join(elements)})"
        elif isinstance(py_obj, (datetime.datetime, datetime.date, datetime.time, UUID)):
            return self.format_quote(as_text(str(py_obj)), is_copy_data)
        else:
            if is_copy_data:
                return str(py_obj)
            else:
                raise TypeError(
                    f"Cannot convert {type(py_obj)} type object to an SQL string. "
                    f"Please register a new adapter for this type via the "
                    f"Cursor.register_sql_literal_adapter() function."
                )

    # noinspection PyArgumentList
    def format_operation_with_parameters(
        self,
        operation: str,
        parameters: Optional[
            Union[List[Any], Tuple[Any], Dict[str, Any]]
        ] = None,
        is_copy_data: bool = False
    ) -> str:
        operation = as_text(operation)

        if isinstance(parameters, dict):
            for key, param in six.iteritems(parameters):
                if not isinstance(key, string_types):
                    key = str(key)
                key = as_text(key)

                value = self.object_to_string(param, is_copy_data)

                # Using a regex with word boundary to correctly handle params with similar names
                # such as :s and :start
                match_str = u":{0}\\b".format(key)
                operation = re.sub(match_str, lambda _: value, operation, flags=re.U)

        elif isinstance(parameters, (tuple, list)):
            tlist = []
            for param in parameters:
                value = self.object_to_string(param, is_copy_data)
                tlist.append(value)

            operation = operation % tuple(tlist)
        else:
            raise TypeError("Argument 'parameters' must be dict or tuple/list")

        return operation

    @staticmethod
    def format_quote(param: str, is_copy_data: bool) -> str:
        if is_copy_data:
            return u'"{0}"'.format(re.escape(param))
        else:
            return u"'{0}'".format(param.replace(u"'", u"''"))

    async def _execute_simple_query(self, query: str) -> None:
        """
        Send the query to the server using the simple query protocol.
        Return True if this query contained no SQL (e.g. the string "--comment")
        """
        self._logger.info(u'Execute simple query: [{}]'.format(query))

        # All of the statements in the query are sent here in a single message
        await self.connection.write(messages.Query(query))

        # The first response could be a number of things:
        #   ErrorResponse: Something went wrong on the server.
        #   EmptyQueryResponse: The query being executed is empty.
        #   RowDescription: This is the "normal" case when executing a query.
        #                   It marks the start of the results.
        #   CommandComplete: This occurs when executing DDL/transactions.
        self._message = await self.connection.read_message()
        if isinstance(self._message, messages.ErrorResponse):
            raise errors.QueryError.from_error_response(self._message, query)
        elif isinstance(self._message, messages.RowDescription):
            self.description = [Column(fd, self.unicode_error) for fd in self._message.fields]
            self._message = await self.connection.read_message()
            if isinstance(self._message, messages.ErrorResponse):
                raise errors.QueryError.from_error_response(self._message, query)
            elif isinstance(self._message, messages.VerifyFiles):
                await self._handle_copy_local_protocol()

    async def _handle_copy_local_protocol(self) -> None:
        if self.connection.options['disable_copy_local']:
            msg = 'COPY LOCAL operation is disabled.'
            await self.connection.write(messages.CopyError(msg))
            await self.flush_to_query_ready()
            raise errors.InterfaceError(msg)

        # Extract info from VerifyFiles message
        input_files = self._message.input_files
        rejections_file = self._message.rejections_file
        exceptions_file = self._message.exceptions_file

        # Verify the file(s) present in the COPY FROM LOCAL statement are indeed accessible
        self.valid_write_file_path = []
        try:
            # Check that the output files are writable
            if rejections_file != '':
                if rejections_file not in self.operation:
                    raise errors.MessageError(
                        f'Server requests for writing to'
                        f' invalid rejected file path: {rejections_file}'
                    )
                os_utils.check_file_writable(rejections_file)
                self.valid_write_file_path.append(rejections_file)
            if exceptions_file != '':
                if exceptions_file not in self.operation:
                    raise errors.MessageError(
                        f'Server requests for writing to'
                        f' invalid exceptions file path: {exceptions_file}'
                    )
                os_utils.check_file_writable(exceptions_file)
                self.valid_write_file_path.append(exceptions_file)

            # Check that the input files are readable
            self.valid_read_file_path = self._check_copy_local_files(input_files)

            await self.connection.write(
                messages.VerifiedFiles(self.valid_read_file_path)
            )
        except Exception as e:
            tb = sys.exc_info()[2]
            stk = traceback.extract_tb(tb, 1)
            await self.connection.write(messages.CopyError(str(e), stk[0]))
            await self.flush_to_query_ready()
            raise

        # Server should be ready to receive copy data from STDIN/files
        self._message = await self.connection.read_message()
        if isinstance(self._message, messages.ErrorResponse):
            raise errors.QueryError.from_error_response(self._message, self.operation)
        elif not isinstance(self._message, (messages.CopyInResponse, messages.LoadFile)):
            raise errors.MessageError(
                f'Unexpected COPY FROM LOCAL state: {type(self._message).__name__}'
            )
        try:
            if isinstance(self._message, messages.CopyInResponse):
                self._logger.info('Sending STDIN data to server')
                if len(self.copy_stdin_list) == 0:
                    raise ValueError(
                        'No STDIN source to load.'
                        ' Please specify "copy_stdin" parameter in Cursor.execute()'
                    )
                stdin = self.copy_stdin_list.pop(0)
                await self._send_copy_data(stdin, self.buffer_size)
                await self.connection.write(messages.EndOfBatchRequest())
                await self._read_copy_data_response(is_stdin_copy=True)
            elif isinstance(self._message, messages.LoadFile):
                while True:
                    await self._send_copy_file_data()
                    if not await self._read_copy_data_response():
                        break
        except errors.QueryError:
            # A server-detected error.
            # The server issues an ErrorResponse message and a ReadyForQuery message.
            raise
        except Exception as e:
            # A client-detected error.
            # The client terminates COPY LOCAL protocol by sending a CopyError message,
            # which will cause the COPY SQL statement to fail with an ErrorResponse message.
            tb = sys.exc_info()[2]
            stk = traceback.extract_tb(tb, 1)
            await self.connection.write(messages.CopyError(str(e), stk[0]))
            await self.flush_to_query_ready()
            raise

    def _check_copy_local_files(self, input_files: List[str]) -> List[str]:
        # Return an empty list when the copy input is STDIN
        if len(input_files) == 0:
            return []

        file_list = []
        for file_pattern in input_files:
            if file_pattern not in self.operation:
                raise errors.MessageError(
                    f'Server requests for loading invalid'
                    f' file: {file_pattern}, Query: {self.operation}'
                )
            # Expand the glob patterns
            expanded_files = glob.glob(file_pattern)
            if len(expanded_files) == 0:
                raise OSError(f'{file_pattern} does not exist')
            # Check file permissions
            for f in expanded_files:
                os_utils.check_file_readable(f)
                file_list.append(f)
        # Return a non-empty list when the copy input is FILE
        # Note: Sending an empty list of files will make server kill the session.
        return file_list

    async def _send_copy_data(
        self,
        # TODO: May be not asyncio.StreamReader type...
        stream: asyncio.StreamReader,
        buffer_size: int
    ) -> None:
        # Send zero or more CopyData messages, forming a stream of input data
        while True:
            chunk = stream.read(buffer_size)
            if not chunk:
                break
            await self.connection.write(messages.CopyData(chunk, self.unicode_error))

    async def _send_copy_file_data(self) -> None:
        filename = self._message.filename
        self._logger.info(f'Sending {filename} data to server')

        if filename not in self.valid_read_file_path:
            raise errors.MessageError(
                f'Server requests for loading invalid'
                f' file: {filename}'
            )

        # TODO: Rewrite to aiofiles
        with open(filename, "rb") as f:
            await self._send_copy_data(f, self.buffer_size)
        await self.connection.write(messages.EndOfBatchRequest())

    async def _read_copy_data_response(self, is_stdin_copy: bool = False) -> bool:
        """Return True if the server wants us to load more data, false if we are done"""
        self._message = await self.connection.read_expected_message(
            END_OF_BATCH_RESPONSES
        )
        # Check for rejections during this load
        while isinstance(self._message, messages.WriteFile):
            if self._message.filename == '':
                self._logger.info(
                    f'COPY-LOCAL rejected row numbers: {self._message.rejected_rows}'
                )
            elif self._message.filename in self.valid_write_file_path:
                self._message.write_to_disk(await self.connection, self.buffer_size)
            else:
                raise errors.MessageError(
                    f'Server requests for writing to'
                    f' invalid file path: {self._message.filename}'
                )
            self._message = await self.connection.read_expected_message(
                END_OF_BATCH_RESPONSES
            )

        # For STDIN copy, there will be no incoming message until we send
        # another EndOfBatchRequest or CopyDone
        if is_stdin_copy:
            await self.connection.write(messages.CopyDone())  # End this copy
            self._message = await self.connection.read_message()
            if isinstance(self._message, messages.ErrorResponse):
                raise errors.QueryError.from_error_response(self._message, self.operation)
            return False

        # For file copy, peek the next message
        self._message = await self.connection.read_message()
        if isinstance(self._message, messages.LoadFile):
            # Indicate there are more local files to load
            return True
        elif isinstance(self._message, messages.ErrorResponse):
            raise errors.QueryError.from_error_response(self._message, self.operation)
        elif not isinstance(self._message, messages.CopyDoneResponse):
            raise errors.MessageError(
                f'Unexpected COPY FROM LOCAL state: {type(self._message).__name__}'
            )
        return False

    async def _error_handler(self, msg: Any) -> None:
        await self.connection.write(messages.Sync())
        raise errors.QueryError.from_error_response(msg, self.operation)

    async def _prepare(self, query: str) -> None:
        """
        Send the query to be prepared to the server. The server will parse the
        query and return some metadata.
        """
        self._logger.info(f'Prepare a statement: [{query}]')

        # Send Parse message to server
        # We don't need to tell the server the parameter types yet
        await self.connection.write(
            messages.Parse(self.prepared_name, query, param_types=())
        )
        # Send Describe message to server
        await self.connection.write(
            messages.Describe('prepared_statement', self.prepared_name)
        )
        await self.connection.write(messages.Flush())

        # Read expected message: ParseComplete
        self._message = await self.connection.read_expected_message(
            messages.ParseComplete, self._error_handler
        )

        # Read expected message: ParameterDescription
        self._message = await self.connection.read_expected_message(
            messages.ParameterDescription, self._error_handler
        )
        self._param_metadata = self._message.parameters

        # Read expected message: RowDescription or NoData
        self._message = await self.connection.read_expected_message(
            (messages.RowDescription, messages.NoData),
            self._error_handler
        )
        if isinstance(self._message, messages.NoData):
            self.description = None  # response was NoData for a DDL/transaction PreparedStatement
        else:
            self.description = [
                Column(fd, self.unicode_error)
                for fd in self._message.fields
            ]

        # Read expected message: CommandDescription
        self._message = await self.connection.read_expected_message(
            messages.CommandDescription, self._error_handler
        )
        if len(self._message.command_tag) == 0:
            msg = 'The statement being prepared is empty'
            self._logger.error(msg)
            await self.connection.write(messages.Sync())
            raise errors.EmptyQueryError(msg)

        self._logger.info('Finish preparing the statement')

    async def _execute_prepared_statement(
        self,
        list_of_parameter_values: List[Any]
    ):
        """
        Send multiple statement parameter sets to the server using the extended
        query protocol. The server would bind and execute each set of parameter
        values.

        This function should not be called without first calling _prepare() to
        prepare a statement.
        """
        portal_name = ""
        parameter_type_oids = [metadata['data_type_oid'] for metadata in self._param_metadata]
        parameter_count = len(self._param_metadata)

        try:
            if len(list_of_parameter_values) == 0:
                raise ValueError("Empty list/tuple, nothing to execute")
            for parameter_values in list_of_parameter_values:
                if parameter_values is None:
                    parameter_values = ()
                self._logger.info(f'Bind parameters: {parameter_values}')
                if len(parameter_values) != parameter_count:
                    msg = ()
                    raise ValueError(
                        f"Invalid number of parameters for {parameter_values}:"
                        f" {len(parameter_values)} given, {parameter_count} expected"
                    )
                await self.connection.write(
                    messages.Bind(
                        portal_name, self.prepared_name, parameter_values,
                        parameter_type_oids
                    )
                )
                await self.connection.write(messages.Execute(portal_name, 0))
            await self.connection.write(messages.Sync())
        except Exception as e:
            self._logger.error(str(e))
            # the server will not send anything until we issue a sync
            await self.connection.write(messages.Sync())
            self._message = await self.connection.read_message()
            raise

        await self.connection.write(messages.Flush())

        # Read expected message: BindComplete
        await self.connection.read_expected_message(messages.BindComplete)

        self._message = await self.connection.read_message()
        if isinstance(self._message, messages.ErrorResponse):
            raise errors.QueryError.from_error_response(self._message, self.prepared_sql)

    async def _close_prepared_statement(self) -> None:
        """
        Close the prepared statement on the server.
        """
        self.prepared_sql = None
        await self.flush_to_query_ready()
        await self.connection.write(
            messages.Close('prepared_statement', self.prepared_name)
        )
        await self.connection.write(messages.Flush())
        self._message = await self.connection.read_expected_message(
            messages.CloseComplete
        )
        await self.connection.write(messages.Sync())



# Copyright (c) 2018-2021 Micro Focus or one of its affiliates.
# Copyright (c) 2018 Uber Technologies, Inc.
# Copyright (c) 2021 Ivan Galin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) 2013-2017 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from __future__ import print_function, division, absolute_import

import asyncio
import base64
import logging
import socket
import ssl
import getpass
import types
import uuid
from struct import unpack
from collections import deque, namedtuple

# noinspection PyCompatibility,PyUnresolvedReferences
from typing import (
    Any, Awaitable, Callable, Coroutine, Dict, KeysView, Optional, Tuple, Type, Union,
    List, TYPE_CHECKING
)

from six import raise_from, string_types, integer_types, PY3

if not PY3:
    raise ValueError('Not supported python version. Supports only python3 and high.')

from urllib.parse import urlparse, parse_qs
import aiovertica
from aiovertica import errors, messages
from aiovertica.cursor import Cursor, CursorType
from aiovertica.messages.message import BackendMessage, FrontendMessage
from aiovertica.messages.frontend_messages import CancelRequest
from aiovertica.log import VerticaLogging

DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 5433
DEFAULT_PASSWORD = ''
DEFAULT_AUTOCOMMIT = False
DEFAULT_BACKUP_SERVER_NODE = []
DEFAULT_KRB_SERVICE_NAME = 'vertica'
DEFAULT_LOG_LEVEL = logging.WARNING
DEFAULT_LOG_PATH = 'aiovertica.log'
try:
    DEFAULT_USER = getpass.getuser()
except Exception as e:
    DEFAULT_USER = None
    print("WARN: Cannot get the login user name: {}".format(str(e)))


StreamRWs = Tuple[asyncio.StreamReader, asyncio.StreamWriter]


async def connect(**kwargs: KeysView[Any]) -> 'Connection':
    """Opens a new connection to a Vertica database."""
    conn = Connection(kwargs)
    return await conn()


def parse_dsn(dsn: str) -> Dict[str, Any]:
    """Parse connection string into a dictionary of keywords and values.
       Connection string format:
           vertica://<user>:<password>@<host>:<port>/<database>?k1=v1&k2=v2&...
    """
    url = urlparse(dsn)
    if url.scheme != 'vertica':
        raise ValueError("Only vertica:// scheme is supported.")

    # Ignore blank/invalid values
    result = {k: v for k, v in (
        ('host', url.hostname),
        ('port', url.port),
        ('user', url.username),
        ('password', url.password),
        ('database', url.path[1:])) if v
              }
    for key, values in parse_qs(url.query, keep_blank_values=True).items():
        # Try to get the last non-blank value in the list of values for each key
        for i in reversed(range(len(values))):
            value = values[i]
            if value != '':
                break

        if value == '' and key != 'log_path':
            # blank values are to be ignored
            continue
        elif key == 'backup_server_node':
            continue
        elif key in ('connection_load_balance', 'use_prepared_statements',
                     'disable_copy_local', 'ssl', 'autocommit'):
            lower = value.lower()
            if lower in ('true', 'on', '1'):
                result[key] = True
            elif lower in ('false', 'off', '0'):
                result[key] = False
        elif key == 'connection_timeout':
            result[key] = float(value)
        elif key == 'log_level' and value.isdigit():
            result[key] = int(value)
        else:
            result[key] = value

    return result

_AddressEntry = namedtuple('_AddressEntry', ['host', 'resolved', 'data'])

class _AddressList(object):
    def __init__(self, host, port, backup_nodes, logger):
        """Creates a new deque with the primary host first, followed by any backup hosts"""

        self._logger = logger

        # Items in address_deque are _AddressEntry values.
        #   host is the original hostname/ip, used by SSL option check_hostname
        #   - when resolved is False, data is port
        #   - when resolved is True, data is the 5-tuple from socket.getaddrinfo
        # This allows for lazy resolution. Seek peek() for more.
        self.address_deque = deque()

        # load primary host into address_deque
        self._append(host, port)

        # load backup nodes into address_deque
        if not isinstance(backup_nodes, list):
            err_msg = 'Connection option "backup_server_node" must be a list'
            self._logger.error(err_msg)
            raise TypeError(err_msg)

        # Each item in backup_nodes should be either
        # a host name or IP address string (using default port) or
        # a (host, port) tuple
        for node in backup_nodes:
            if isinstance(node, string_types):
                self._append(node, DEFAULT_PORT)
            elif isinstance(node, tuple) and len(node) == 2:
                self._append(node[0], node[1])
            else:
                err_msg = ('Each item of connection option "backup_server_node"'
                           ' must be a host string or a (host, port) tuple')
                self._logger.error(err_msg)
                raise TypeError(err_msg)

        self._logger.debug('Address list: {0}'.format(list(self.address_deque)))

    def _append(self, host, port):
        if not isinstance(host, string_types):
            err_msg = 'Host must be a string: invalid value: {0}'.format(host)
            self._logger.error(err_msg)
            raise TypeError(err_msg)

        if not isinstance(port, (string_types, integer_types)):
            err_msg = 'Port must be an integer or a string: invalid value: {0}'.format(port)
            self._logger.error(err_msg)
            raise TypeError(err_msg)
        elif isinstance(port, string_types):
            try:
                port = int(port)
            except ValueError as e:
                err_msg = 'Port "{0}" is not a valid string: {1}'.format(port, e)
                self._logger.error(err_msg)
                raise ValueError(err_msg)

        if port < 0 or port > 65535:
            err_msg = 'Invalid port number: {0}'.format(port)
            self._logger.error(err_msg)
            raise ValueError(err_msg)

        self.address_deque.append(_AddressEntry(host=host, resolved=False, data=port))

    def push(self, host, port):
        self.address_deque.appendleft(_AddressEntry(host=host, resolved=False, data=port))

    def pop(self):
        self.address_deque.popleft()

    def peek(self):
        # do lazy DNS resolution, returning the leftmost socket.getaddrinfo result
        if len(self.address_deque) == 0:
            return None

        while len(self.address_deque) > 0:
            self._logger.debug('Peek at address list: {0}'.format(list(self.address_deque)))
            entry = self.address_deque[0]
            if entry.resolved:
                # return a resolved sockaddrinfo
                return entry.data
            else:
                # DNS resolve a single host name to multiple IP addresses
                self.address_deque.popleft()
                host, port = entry.host, entry.data
                try:
                    resolved_hosts = socket.getaddrinfo(host, port, 0, socket.SOCK_STREAM)
                except Exception as e:
                    self._logger.warning('Error resolving host "{0}" on port {1}: {2}'.format(host, port, e))
                    continue

                # add resolved addrinfo (AF_INET and AF_INET6 only) to deque
                for addrinfo in reversed(resolved_hosts):
                    if addrinfo[0] in (socket.AF_INET, socket.AF_INET6):
                        self.address_deque.appendleft(_AddressEntry(
                            host=host, resolved=True, data=addrinfo))
        return None

    def peek_host(self):
        # returning the leftmost host result
        self._logger.debug('Peek host at address list: {0}'.format(list(self.address_deque)))
        if len(self.address_deque) == 0:
            return None
        return self.address_deque[0].host


def _generate_session_label() -> str:
    """Return generated session label"""
    return f'vertica-python-{aiovertica.__version__}-{uuid.uuid1()}'


class Connection:
    def __init__(self, options: Optional[Dict[str, Any]] = None):
        self.parameters = {}
        self.session_id = None
        self.backend_pid = None
        self.backend_key = None
        self.transaction_status = None
        self.reader: Optional[asyncio.StreamReader] = None
        self.writer: Optional[asyncio.StreamWriter] = None

        options = options or {}
        self.options = parse_dsn(options['dsn']) if 'dsn' in options else {}
        self.options.update({key: value for key, value in options.items() \
                             if key == 'log_path' or (key != 'dsn' and value is not None)})

        # Set up connection logger
        logger_name = f'vertica_{id(self)}_{uuid.uuid4()}'  # must be a unique value
        self._logger = logging.getLogger(logger_name)

        if 'log_level' not in self.options and 'log_path' not in self.options:
            # logger is disabled by default
            self._logger.disabled = True
        else:
            self.options.setdefault('log_level', DEFAULT_LOG_LEVEL)
            self.options.setdefault('log_path', DEFAULT_LOG_PATH)
            VerticaLogging.setup_logging(logger_name, self.options['log_path'],
                self.options['log_level'], id(self))

        self.options.setdefault('host', DEFAULT_HOST)
        self.options.setdefault('port', DEFAULT_PORT)
        if 'user' not in self.options:
            if DEFAULT_USER:
                self.options['user'] = DEFAULT_USER
            else:
                msg = 'Connection option "user" is required'
                self._logger.error(msg)
                raise KeyError(msg)
        self.options.setdefault('database', self.options['user'])
        self.options.setdefault('password', DEFAULT_PASSWORD)
        self.options.setdefault('autocommit', DEFAULT_AUTOCOMMIT)
        self.options.setdefault('session_label', _generate_session_label())
        self.options.setdefault('backup_server_node', DEFAULT_BACKUP_SERVER_NODE)
        self.options.setdefault('kerberos_service_name', DEFAULT_KRB_SERVICE_NAME)
        # Kerberos authentication hostname defaults to the host value here so
        # the correct value cannot be overwritten by load balancing or failover
        self.options.setdefault('kerberos_host_name', self.options['host'])

        self.address_list = _AddressList(self.options['host'], self.options['port'],
            self.options['backup_server_node'], self._logger)

        # we only support one cursor per connection
        self.options.setdefault('unicode_error', None)
        self._cursor = Cursor(self, self._logger, cursor_type=None,
            unicode_error=self.options['unicode_error'])

        # knob for using server-side prepared statements
        self.options.setdefault('use_prepared_statements', False)
        is_connection_prepared_statements = (
            'enabled' if self.options['use_prepared_statements'] else 'disabled'
        )
        self._logger.debug(
            f'Connection prepared statements is {is_connection_prepared_statements}'
        )

        # knob for disabling COPY LOCAL operations
        self.options.setdefault('disable_copy_local', False)
        is_disable_copy_local = (
            'disabled' if self.options['disable_copy_local'] else 'enabled'
        )
        self._logger.debug(f'COPY LOCAL operation is {is_disable_copy_local}')

    async def __call__(self) -> 'Connection':
        self._logger.info(
            f'Connecting as user "{self.options["user"]}" '
            f'to database "{self.options["database"]}" '
            f'on host "{self.options["host"]}" with port "{self.options["port"]}"'
        )

        await self.startup_connection()

        # Set autocommit settings (need connection)
        # Initially, for a new session, autocommit is off
        if self.options['autocommit']:
            await self.set_autocommit(True)

        self._logger.info('Connection is ready')
        return self

    #############################################
    # supporting `async with` statements
    #############################################
    async def __aenter__(self) -> 'Connection':
        return await self.__call__()

    async def __aexit__(
        self,
        type_: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[types.TracebackType]
    ) -> None:
        await self.close()

    #############################################
    # dbapi methods
    #############################################
    async def close(self) -> None:
        self._logger.info('Close the connection')
        try:
            await self.write(messages.Terminate())
        finally:
            await self.close_socket()

    async def commit(self) -> None:
        if self.closed():
            raise errors.ConnectionError('Connection is closed')

        cur = self.cursor()
        await cur.execute('COMMIT;')

    async def rollback(self) -> None:
        if self.closed():
            raise errors.ConnectionError('Connection is closed')

        cur = self.cursor()
        await cur.execute('ROLLBACK;')

    def cursor(
        self, cursor_type: Optional[CursorType] = None
    ) -> Cursor:
        if self.closed():
            raise errors.ConnectionError('Connection is closed')

        if self._cursor.closed():
            self._cursor._closed = False

        # let user change type if they want?
        self._cursor.cursor_type = cursor_type
        return self._cursor

    #############################################
    # non-dbapi methods
    #############################################
    @property
    def autocommit(self):
        """Read the connection's AUTOCOMMIT setting from cache"""
        return self.parameters.get('auto_commit', 'off') == 'on'

    @autocommit.setter
    def autocommit(self, value) -> None:
        raise NotImplementedError(
            "Python autocommit setter is not available, since it cannot be asynchronous."
            " Please use `await self.set_autocommit(value)` instead."
        )

    async def set_autocommit(self, value: bool) -> None:
        """Change the connection's AUTOCOMMIT setting"""
        if self.autocommit is value:
            return
        val = 'on' if value else 'off'
        cur = self.cursor()
        await cur.execute(
            f'SET SESSION AUTOCOMMIT TO {val}',
            use_prepared_statements=False
        )
        await cur.fetchall()   # check for errors and update the cache
        self.__dict__['autocommit'] = value

    async def cancel(self) -> None:
        """Cancel the current database operation. This can be called from a
           different thread than the one currently executing a database operation.
        """
        if self.closed():
            raise errors.ConnectionError('Connection is closed')
        self._logger.info('Canceling the current database operation')
        # Must create a new socket connection to the server
        _, writer = await self.establish_socket_connection(
            self.address_list
        )
        await self.write(CancelRequest(self.backend_pid, self.backend_key), writer)
        writer.close()
        await writer.wait_closed()

        self._logger.info('Cancel request issued')

    def opened(self) -> bool:
        return (self.writer is not None
                and self.backend_pid is not None
                and self.transaction_status is not None)

    def closed(self) -> bool:
        return not self.opened()

    def __str__(self) -> str:
        safe_options = {
            key: value
            for key, value in self.options.items()
            if key != 'password'
        }

        return f'<AIOVertica.Connection:{id(self)} {self.parameters=}' \
               f' {self.backend_pid=}, {self.backend_key=},' \
               f' {self.transaction_status=},' \
               f' socket={self._socket()}, options={safe_options}>'

    #############################################
    # internal
    #############################################
    def _socket(self) -> Optional[socket.socket]:
        if not self.writer:
            return None
        _socket: socket.socket = self.writer.get_extra_info("socket")
        return _socket

    def reset_values(self) -> None:
        self.parameters = {}
        self.session_id = None
        self.backend_pid = None
        self.backend_key = None
        self.transaction_status = None
        self.reader = None
        self.writer = None
        self.address_list = _AddressList(self.options['host'], self.options['port'],
            self.options['backup_server_node'], self._logger)

    async def _writer(self) -> asyncio.StreamWriter:
        if self.writer:
            return self.writer

        # the initial establishment of the client connection
        reader, writer = await self.establish_socket_connection(self.address_list)

        # enable load balancing
        load_balance_options = self.options.get('connection_load_balance')
        is_load_balance = 'enabled' if load_balance_options else 'disabled'
        self._logger.debug(f'Connection load balance option is {is_load_balance}')
        if load_balance_options:
            reader, writer = await self.balance_load(reader, writer)

        # enable SSL
        ssl_options = self.options.get('ssl')
        is_ssl_enabled = 'enabled' if ssl_options else 'disabled'
        self._logger.debug(f'SSL option is {is_ssl_enabled}')
        if ssl_options:
            reader, writer = await self.enable_ssl(reader, writer, ssl_options)

        self.reader = reader
        self.writer = writer

        return self.writer

    async def balance_load(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> StreamRWs:
        # Send load balance request and read server response
        self._logger.debug(f'=> {messages.LoadBalanceRequest()}')

        writer.write(messages.LoadBalanceRequest().get_message())
        await writer.drain()

        response = await reader.read(1)

        if response == b'Y':
            size = unpack('!I', await reader.read(4))[0]
            if size < 4:
                err_msg = f"Bad message size: {size}"
                self._logger.error(err_msg)
                raise errors.MessageError(err_msg)
            res = BackendMessage.from_type(
                type_=response, data=await reader.read(size - 4)
            )
            self._logger.debug(f'<= {res}')
            host = res.get_host()
            port = res.get_port()
            self._logger.info(
                f'Load balancing to host "{host}" on port {port}'
            )

            peer = writer.transport.get_extra_info('peername')
            if peer:
                socket_host, socket_port = peer[0], peer[1]
                if host == socket_host and port == socket_port:
                    self._logger.info(
                        f'Already connecting to host "{host}" on port {port}.'
                        f' Ignore load balancing.'
                    )
                    return reader, writer

            # Push the new host onto the address list before connecting again. Note that this
            # will leave the originally-specified host as the first failover possibility.
            self.address_list.push(host, port)
            writer.close()
            await writer.wait_closed()
            reader, writer = self.establish_socket_connection(self.address_list)
        else:
            self._logger.debug(f'<= LoadBalanceResponse: {response}')
            self._logger.warning("Load balancing requested but not supported by server")

        return reader, writer

    async def enable_ssl(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        ssl_options: Union[Dict[str, Any], ssl.SSLContext],
    ) -> StreamRWs:
        # Send SSL request and read server response
        self._logger.debug('=> %s', messages.SslRequest())
        writer.write(messages.SslRequest().get_message())
        await writer.drain()
        response = await reader.read(1)
        self._logger.debug(f'<= SslResponse: {response}')
        if response == b'S':
            self._logger.info('Enabling SSL')
            try:
                loop = asyncio.get_running_loop()
                if isinstance(ssl_options, ssl.SSLContext):
                    server_host = self.address_list.peek_host()
                    if server_host is None:   # This should not happen
                        msg = 'Cannot get the connected server host while enabling SSL'
                        self._logger.error(msg)
                        raise errors.ConnectionError(msg)
                    transport = await loop.start_tls(
                        writer.transport,
                        writer.transport.get_protocol(),
                        sslcontext=ssl_options,
                        server_hostname=server_host,
                    )
                    writer._transport = transport
                    reader.set_transport(transport)
                else:
                    transport = await loop.start_tls(
                        writer.transport,
                        writer.transport.get_protocol(),
                        sslcontext=ssl.SSLContext(ssl.PROTOCOL_SSLv23)
                    )
                    writer._transport = transport
                    reader.set_transport(transport)
            except ssl.CertificateError as e:
                raise_from(errors.ConnectionError(str(e)), e)
            except ssl.SSLError as e:
                raise_from(errors.ConnectionError(str(e)), e)
        else:
            err_msg = "SSL requested but not supported by server"
            self._logger.error(err_msg)
            raise errors.SSLNotSupported(err_msg)

        self.writer = writer
        self.reader = reader
        return reader, writer

    async def establish_socket_connection(
        self,
        address_list: _AddressList
    ) -> StreamRWs:
        """
        Given a list of database node addresses, establish the socket
        connection to the database server. Return a connected socket object.
        """
        addrinfo = address_list.peek()
        writer = None
        last_exception = None

        # Failover: loop to try all addresses
        while addrinfo:
            (family, socktype, proto, canonname, sockaddr) = addrinfo
            last_exception = None

            # _AddressList filters all addrs to AF_INET and AF_INET6, which both
            # have host and port as values 0, 1 in the sockaddr tuple.
            host = sockaddr[0]
            port = sockaddr[1]

            self._logger.info(
                f'Establishing connection to host "{host}" on port {port}'
            )

            try:
                reader, writer = await asyncio.open_connection(
                    host=self.options['host'],
                    port=self.options['port'],
                )

                self.reader = reader
                self.writer = writer

                break
            except Exception as e:
                self._logger.info(
                    f'Failed to connect to host "{host}" on port {port}: {e}'
                )
                last_exception = e
                address_list.pop()
                addrinfo = address_list.peek()
                if writer:
                    writer.close()
                    await writer.wait_closed()

        # all of the addresses failed
        if (self.reader is None or self.writer is None) or last_exception:
            err_msg = 'Failed to establish a connection to the primary server or any backup address.'
            self._logger.error(err_msg)
            raise errors.ConnectionError(err_msg)

        return self.reader, self.writer

    def ssl(self) -> bool:
        return self.writer.get_extra_info('ssl_object') is not None

    async def write(
        self,
        message: FrontendMessage,
        writer: Optional[asyncio.StreamWriter] = None
    ) -> None:
        if not isinstance(message, FrontendMessage):
            raise TypeError(f"invalid message: ({message})")
        if writer is None:
            writer = await self._writer()

        self._logger.debug(f'=> {message}')
        try:
            for data in message.fetch_message():
                try:
                    writer.write(data)
                    await writer.drain()
                except Exception:
                    self._logger.error("couldn't send message")
                    raise

        except Exception as e:
            self.writer.close()
            await self.writer.wait_closed()
            self._logger.error(str(e))
            if isinstance(e, IOError):
                raise_from(errors.ConnectionError(str(e)), e)
            else:
                raise

    async def close_socket(self) -> None:
        try:
            if self.writer is not None:
                self.writer.close()
                await self.writer.wait_closed()
        finally:
            self.reset_values()

    async def reset_connection(self) -> None:
        await self.close_socket()
        await self.startup_connection()

    @staticmethod
    def is_asynchronous_message(
        message: Union[BackendMessage, FrontendMessage]
    ) -> bool:
        # Check if it is an asynchronous response message
        # Note: ErrorResponse is a subclass of NoticeResponse
        return (
            isinstance(message, messages.ParameterStatus)
            or (
                isinstance(message, messages.NoticeResponse)
                and not isinstance(message, messages.ErrorResponse)
            )
        )

    def handle_asynchronous_message(
        self, message: Union[BackendMessage, FrontendMessage]
    ) -> None:
        if isinstance(message, messages.ParameterStatus):
            if message.name == 'protocol_version':
                message.value = int(message.value)
            self.parameters[message.name] = message.value
        elif (isinstance(message, messages.NoticeResponse) and
              not isinstance(message, messages.ErrorResponse)):
            if getattr(self, 'notice_handler', None) is not None:
                self.notice_handler(message)
            else:
                self._logger.warning(message.error_message())

    async def read_string(self) -> bytearray:
        s = bytearray()
        while True:
            char = await self.read_bytes(1)
            if char == b'\x00':
                break
            s.extend(char)
        return s

    async def read_message(self) -> BackendMessage:
        while True:
            try:
                type_ = await self.read_bytes(1)
                size = unpack('!I', await self.read_bytes(4))[0]
                if size < 4:
                    raise errors.MessageError(f"Bad message size: {size}")
                if type_ == messages.WriteFile.message_id:
                    # The whole WriteFile message may not be read at here.
                    # Instead, only the file name and file length is read.
                    # This is because the message could be too large to read all at once.
                    f = await self.read_string()
                    filename = f.decode('utf-8')
                    file_length = unpack('!I', await self.read_bytes(4))[0]
                    size -= 4 + len(f) + 1 + 4
                    if size != file_length:
                        raise errors.MessageError(f"Bad message size: {size}")
                    if filename == '':
                        # If there is no filename, then this is really RETURNREJECTED data, not a rejected file
                        if file_length % 8 != 0:
                            raise errors.MessageError(
                                f"Bad RETURNREJECTED data size: {file_length}"
                            )
                        data = await self.read_bytes(file_length)
                        message = messages.WriteFile(filename, file_length, data)
                    else:
                        # The rest of the message is read later with write_to_disk()
                        message = messages.WriteFile(filename, file_length)
                else:
                    message = BackendMessage.from_type(
                        type_, await self.read_bytes(size - 4)
                    )
                self._logger.debug(f'<= {message}')
                self.handle_asynchronous_message(message)
                # handle transaction status
                if isinstance(message, messages.ReadyForQuery):
                    self.transaction_status = message.transaction_status
            except (SystemError, IOError) as e:
                await self.close_socket()
                # noinspection PyTypeChecker
                self._logger.error(e)
                raise_from(errors.ConnectionError(str(e)), e)
            if not self.is_asynchronous_message(message):
                break
        return message

    async def read_expected_message(
        self,
        expected_types: Union[messages.BindComplete, Tuple[messages.BindComplete]],
        error_handler: Optional[
            Callable[[Any], Coroutine[Any, Any, None]]
        ] = None
    ) -> None:
        # Reads a message and does some basic error handling.
        # expected_types must be a class (e.g. messages.BindComplete) or a tuple of classes
        message = self.read_message()
        if isinstance(message, expected_types):
            return message
        elif isinstance(message, messages.ErrorResponse):
            if error_handler is not None:
                await error_handler(message)
            else:
                raise errors.DatabaseError(message.error_message())
        else:
            msg = f'Received unexpected message type: {type(message)}.'
            if isinstance(expected_types, tuple):
                stringify_expected_types = ", ".join(
                    [t.__name__ for t in expected_types]
                )
                msg += f'Expected types: {stringify_expected_types}'
            else:
                msg += f'Expected type: {expected_types.__name__}'
            self._logger.error(msg)
            raise errors.MessageError(msg)

    async def read_bytes(self, n: int) -> bytes:
        if n == 1:
            result = await self.reader.read(1)
            if not result:
                raise errors.ConnectionError("Connection closed by Vertica")
        else:
            result = await self.reader.readexactly(n)
            if len(result) == 0:
                raise errors.ConnectionError("Connection closed by Vertica")
        return result

    async def send_GSS_response_and_receive_challenge(self, response):
        # Send the GSS response data to the vertica server
        token = base64.b64decode(response)
        await self.write(messages.Password(token, messages.Authentication.GSS))
        # Receive the challenge from the vertica server
        message = self.read_expected_message(messages.Authentication)
        if message.code != messages.Authentication.GSS_CONTINUE:
            msg = ('Received unexpected message type: Authentication(type={}).'
                   ' Expected type: Authentication(type={})'.format(
                message.code, messages.Authentication.GSS_CONTINUE))
            self._logger.error(msg)
            raise errors.MessageError(msg)
        return message.auth_data

    def make_GSS_authentication(self):
        try:
            import kerberos
        except ImportError as e:
            raise errors.ConnectionError("{}\nCannot make a Kerberos "
                                         "authentication because no Kerberos package is installed. "
                                         "Get it with 'pip install kerberos'.".format(str(e)))

        # Set GSS flags
        gssflag = (kerberos.GSS_C_DELEG_FLAG | kerberos.GSS_C_MUTUAL_FLAG |
                   kerberos.GSS_C_SEQUENCE_FLAG | kerberos.GSS_C_REPLAY_FLAG)

        # Generate the GSS-style service principal name
        service_principal = "{}@{}".format(self.options['kerberos_service_name'],
            self.options['kerberos_host_name'])

        # Initializes a context object with a service principal
        self._logger.info('Initializing a context for GSSAPI client-side '
                          'authentication with service principal {}'.format(service_principal))
        try:
            result, context = kerberos.authGSSClientInit(service_principal, gssflags=gssflag)
        except kerberos.GSSError as err:
            msg = "GSSAPI initialization error: {}".format(str(err))
            self._logger.error(msg)
            raise errors.KerberosError(msg)
        if result != kerberos.AUTH_GSS_COMPLETE:
            msg = ('Failed to initialize a context for GSSAPI client-side '
                   'authentication with service principal {}'.format(service_principal))
            self._logger.error(msg)
            raise errors.KerberosError(msg)

        # Processes GSSAPI client-side steps
        try:
            challenge = b''
            while True:
                self._logger.info('Processing a single GSSAPI client-side step')
                challenge = base64.b64encode(challenge).decode("utf-8")
                result = kerberos.authGSSClientStep(context, challenge)

                if result == kerberos.AUTH_GSS_COMPLETE:
                    self._logger.info('Result: GSSAPI step complete')
                    break
                elif result == kerberos.AUTH_GSS_CONTINUE:
                    self._logger.info('Result: GSSAPI step continuation')
                    # Get the response from the last successful GSSAPI client-side step
                    response = kerberos.authGSSClientResponse(context)
                    challenge = self.send_GSS_response_and_receive_challenge(response)
                else:
                    msg = "GSSAPI client-side step error status {}".format(result)
                    self._logger.error(msg)
                    raise errors.KerberosError(msg)
        except kerberos.GSSError as err:
            msg = "GSSAPI client-side step error: {}".format(str(err))
            self._logger.error(msg)
            raise errors.KerberosError(msg)

    async def startup_connection(self) -> None:
        user = self.options['user']
        database = self.options['database']
        session_label = self.options['session_label']
        os_user_name = DEFAULT_USER if DEFAULT_USER else ''
        password = self.options['password']

        await self.write(messages.Startup(user, database, session_label, os_user_name))

        while True:
            message = await self.read_message()

            if isinstance(message, messages.Authentication):
                if message.code == messages.Authentication.OK:
                    self._logger.info(
                        f"User {self.options['user']} successfully authenticated"
                    )
                elif message.code == messages.Authentication.CHANGE_PASSWORD:
                    msg = f"The password for user {self.options['user']} has expired"
                    self._logger.error(msg)
                    raise errors.ConnectionError(msg)
                elif message.code == messages.Authentication.PASSWORD_GRACE:
                    self._logger.warning(
                        f"The password for user {self.options['user']}"
                        f" will expire soon. Please consider changing it."
                    )
                elif message.code == messages.Authentication.GSS:
                    self.make_GSS_authentication()
                else:
                    await self.write(
                        messages.Password(
                            password,
                            message.code,
                            {
                                'user': user,
                                'salt': getattr(message, 'salt', None),
                                'usersalt': getattr(message, 'usersalt', None)
                            }
                        )
                    )
            elif isinstance(message, messages.BackendKeyData):
                self.backend_pid = message.pid
                self.backend_key = message.key
            elif isinstance(message, messages.ReadyForQuery):
                break
            elif isinstance(message, messages.ErrorResponse):
                self._logger.error(message.error_message())
                raise errors.ConnectionError(message.error_message())
            else:
                msg = f"Received unexpected startup message: {message}"
                self._logger.error(msg)
                raise errors.MessageError(msg)



# Copyright (c) 2020-2021 Micro Focus or one of its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
WriteFile message

In the copy-local protocol, the server may send a WriteFile message when it
receives a batch of copy data. If the COPY FROM LOCAL command uses the
REJECTED DATA and/or EXCEPTIONS parameters, this message contains content of
rejected rows or exceptions output files. If the command uses the
RETURNREJECTED parameters instead, this message is a series of row numbers
saying which rows in the load were rejected.
"""

from __future__ import print_function, division, absolute_import

from struct import unpack_from

from ..message import BackendMessage


class WriteFile(BackendMessage):
    message_id = b'O'

    def __init__(self, filename, file_length, data=None):
        BackendMessage.__init__(self)
        self.filename = filename
        self.file_length = file_length

        # Parse RETURNREJECTED data
        if self.filename == '':
            row_count = self.file_length // 8
            # Rejected row numbers come in little endian format
            self.rejected_rows = unpack_from('<{0}Q'.format(row_count), data)

    def write_to_disk(self, connection, buffer_size):
        # Read the rest of the message from wire and write the file
        bytes_left = self.file_length
        with open(self.filename, 'ab') as f:
            pos = 0
            while bytes_left > 0:
                bytes_to_read = min(buffer_size, bytes_left)
                content = connection.read_bytes(bytes_to_read)
                f.write(content)
                bytes_left -= bytes_to_read

    def __str__(self):
        return "WriteFile: Filename = {}, FileLength = {}".format(self.filename, self.file_length)


BackendMessage.register(WriteFile)



# Copyright (c) 2020-2021 Micro Focus or one of its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
VerifyFiles message

VerifyFiles message is sent by the server when the client issues
a COPY FROM LOCAL command. The server parses the file names out
of the command, and sends them back to the client in this message.
The client has to verify that these files exist and are readable
before running the copy.
"""

from __future__ import print_function, division, absolute_import

from struct import unpack_from

from ..message import BackendMessage


class VerifyFiles(BackendMessage):
    message_id = b'F'

    def __init__(self, data):
        BackendMessage.__init__(self)
        files_count = unpack_from('!H', data, 0)[0]
        self.input_files = [None] * files_count
        pos = 2
        for i in range(files_count):
            filename = unpack_from("!{0}sx".format(data.find(b'\x00', pos) - pos), data, pos)[0]
            self.input_files[i] = filename.decode('utf-8')
            pos += len(filename) + 1

        filename = unpack_from("!{0}sx".format(data.find(b'\x00', pos) - pos), data, pos)[0]
        self.rejections_file = filename.decode('utf-8')
        pos += len(filename) + 1

        filename = unpack_from("!{0}sx".format(data.find(b'\x00', pos) - pos), data, pos)[0]
        self.exceptions_file = filename.decode('utf-8')

    def __str__(self):
        return "VerifyFiles: InputFiles = {}, RejectedDataFile = {}, ExceptionsFile = {}".format(
               self.input_files, self.rejections_file, self.exceptions_file)


BackendMessage.register(VerifyFiles)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiovertica',
 'aiovertica.messages',
 'aiovertica.messages.backend_messages',
 'aiovertica.messages.frontend_messages',
 'aiovertica.mixins']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'aiovertica',
    'version': '0.0.2',
    'description': 'Small asynchronous driver for Vertica based on vertica_python',
    'long_description': '# aiovertica\nSmall asynchronous driver for Vertica based on [vertica_python](https://github.com/vertica/vertica-python/).\n\n## WARNING\n\nThis project was written "in haste" literally overnight, so it may not be stable.\nOf course, most of the mechanisms are identical to the "project", but in any case, \ntests have not yet been written to be sure of this at least to some extent.\n\nAt the moment, the project has rewritten the mechanisms responsible for working \nwith the network from a synchronous model to an asynchronous one, as well as \nin the `Connection` and` Cursor` modules, the synchronous methods have been replaced \nby asynchronous ones (where necessary), the code formatting has been slightly improved \nand type annotations have been added (but not always correct).\n\n### Foreword\nI would really like to see a ready-made library, ideally the same as `asyncpg`, \nbut unfortunately it is not there. I would like to express my deep gratitude to the team \nthat worked on the `vertica_python` for writing at least a synchronous driver - this greatly \nsimplified the work. I have tried to keep all the licenses and the list of authors from \nthe original project.\n\nI would also be glad if the guys from `vertica_python` added something similar to the \nofficial driver and we would have a simple opportunity to work with the driver in \nasynchronous mode as well.\n\nPerhaps in the future I will create a PR in the official driver (I have several \nproposals for the architecture), but at the moment it would take too long to implement \nthe "correct" embedding into the existing implementation - the code was needed "yesterday".\n\nI will be glad for any help in the development of the project. The `Issues` and` PRs` \nsections are looking forward to seeing you :).\n\n\n## Installation\n\nTo install vertica-python with pip:\n```bash\n# Latest release version\npip install aiovertica\n\n# Latest commit on master branch\npip install git+https://github.com/i8enn/aiovertica.git@master\n```\n\nTo install vertica-python from source, run the following command from the root directory:\n\n    python setup.py install\n\nSource code for vertica-python can be found at:\n\n    https://github.com/i8enn/aiovertica\n\n#### Using Kerberos authentication\n\n> #### WARNING: `Kerberos` not supported in currently version\n\nvertica-python has optional Kerberos authentication support for Unix-like systems, which requires you to install the [kerberos](https://pypi.org/project/kerberos/) package:\n\n    pip install kerberos\n\nNote that `kerberos` is a python extension module, which means you need to install `python-dev`. The command depends on the package manager and will look like\n\n    sudo [yum|apt-get|etc] install python-dev\n\n## Usage\n\n\n**Create connection**\n\n```python\nimport aiovertica\n\nconn_info = {\'host\': \'127.0.0.1\',\n             \'port\': 5433,\n             \'user\': \'some_user\',\n             \'password\': \'some_password\',\n             \'database\': \'a_database\',\n             # autogenerated session label by default,\n             \'session_label\': \'some_label\',\n             # default throw error on invalid UTF-8 results\n             \'unicode_error\': \'strict\',\n             # SSL is disabled by default\n             \'ssl\': False,\n             # autocommit is off by default\n             \'autocommit\': True,\n             # using server-side prepared statements is disabled by default\n             \'use_prepared_statements\': False,\n             # connection timeout is not enabled by default\n             # 5 seconds timeout for a socket operation (Establishing a TCP connection or read/write operation)\n             \'connection_timeout\': 5}\n\n# simple connection, with manual close\ntry:\n    connection = await aiovertica.connect(**conn_info)\n    # do things\nfinally:\n    await connection.close()\n\n# using with for auto connection closing after usage\nasync with vertica_python.connect(**conn_info) as connection:\n    # do things\n```\n\nYou can pass an `ssl.SSLContext` to `ssl` to customize the SSL connection options. For example,\n\n```python\nimport aiovertica\nimport ssl\n\nssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)\nssl_context.verify_mode = ssl.CERT_REQUIRED\nssl_context.check_hostname = True\nssl_context.load_verify_locations(cafile=\'/path/to/ca_file.pem\')\n\nconn_info = {\'host\': \'127.0.0.1\',\n             \'port\': 5433,\n             \'user\': \'some_user\',\n             \'password\': \'some_password\',\n             \'database\': \'a_database\',\n             \'ssl\': ssl_context}\nconnection = await aiovertica.connect(**conn_info)\n\n```\n\nSee more on SSL options [here](https://docs.python.org/3/library/ssl.html).\n\n> #### WARNING: `Kerberos` not supported in currently version\n\nIn order to use Kerberos authentication, install [dependencies](#using-kerberos-authentication) first, and it is the user\'s responsibility to ensure that an Ticket-Granting Ticket (TGT) is available and valid. Whether a TGT is available can be easily determined by running the `klist` command. If no TGT is available, then it first must be obtained by running the `kinit` command or by logging in. You can pass in optional arguments to customize the authentication. The arguments are `kerberos_service_name`, which defaults to "vertica", and `kerberos_host_name`, which defaults to the value of argument `host`. For example,\n\n```python\nimport aiovertica\n\nconn_info = {\'host\': \'127.0.0.1\',\n             \'port\': 5433,\n             \'user\': \'some_user\',\n             \'password\': \'some_password\',\n             \'database\': \'a_database\',\n             # The service name portion of the Vertica Kerberos principal\n             \'kerberos_service_name\': \'vertica_krb\',\n             # The instance or host name portion of the Vertica Kerberos principal\n             \'kerberos_host_name\': \'vcluster.example.com\'}\n\nasync with aiovertica.connect(**conn_info) as conn:\n    # do things\n```\n\nLogging is disabled by default if you do not pass values to both ```log_level``` and ```log_path```.  The default value of ```log_level``` is logging.WARNING. You can find all levels [here](https://docs.python.org/3/library/logging.html#logging-levels). The default value of ```log_path``` is \'vertica_python.log\', the log file will be in the current execution directory. If ```log_path``` is set to ```\'\'``` (empty string) or ```None```, no file handler is set, logs will be processed by root handlers. For example,\n\n```python\nimport aiovertica\nimport logging\n\n## Example 1: write DEBUG level logs to \'./vertica_python.log\'\nconn_info = {\'host\': \'127.0.0.1\',\n             \'port\': 5433,\n             \'user\': \'some_user\',\n             \'password\': \'some_password\',\n             \'database\': \'a_database\',\n             \'log_level\': logging.DEBUG}\nasync with aiovertica.connect(**conn_info) as connection:\n    # do things\n\n## Example 2: write WARNING level logs to \'./path/to/logs/client.log\'\nconn_info = {\'host\': \'127.0.0.1\',\n             \'port\': 5433,\n             \'user\': \'some_user\',\n             \'password\': \'some_password\',\n             \'database\': \'a_database\',\n             \'log_path\': \'path/to/logs/client.log\'}\nasync with aiovertica.connect(**conn_info) as connection:\n   # do things\n\n## Example 3: write INFO level logs to \'/home/admin/logs/vClient.log\'\nconn_info = {\'host\': \'127.0.0.1\',\n             \'port\': 5433,\n             \'user\': \'some_user\',\n             \'password\': \'some_password\',\n             \'database\': \'a_database\',\n             \'log_level\': logging.INFO,\n             \'log_path\': \'/home/admin/logs/vClient.log\'}\nasync with aiovertica.connect(**conn_info) as connection:\n   # do things\n\n## Example 4: use root handlers to process logs by setting \'log_path\' to \'\' (empty string) \nconn_info = {\'host\': \'127.0.0.1\',\n             \'port\': 5433,\n             \'user\': \'some_user\',\n             \'password\': \'some_password\',\n             \'database\': \'a_database\',\n             \'log_level\': logging.DEBUG,\n             \'log_path\': \'\'}\nasync with aiovertica.connect(**conn_info) as connection:\n    # do things\n```\n\nConnection Failover: Supply a list of backup hosts to ```backup_server_node``` for the client to try if the primary host you specify in the connection parameters (```host```, ```port```) is unreachable. Each item in the list should be either a host string (using default port 5433) or a (host, port) tuple. A host can be a host name or an IP address.\n\n```python\nimport aiovertica\n\nconn_info = {\'host\': \'unreachable.server.com\',\n             \'port\': 888,\n             \'user\': \'some_user\',\n             \'password\': \'some_password\',\n             \'database\': \'a_database\',\n             \'backup_server_node\': [\'123.456.789.123\', \'invalid.com\', (\'10.20.82.77\', 6000)]}\nconnection = await aiovertica.connect(**conn_info)\n```\n\nConnection Load Balancing helps automatically spread the overhead caused by client connections across the cluster by having hosts redirect client connections to other hosts. Both the server and the client need to enable load balancing for it to function. If the server disables connection load balancing, the load balancing request from client will be ignored.\n\n```python\nimport aiovertica\n\nconn_info = {\'host\': \'127.0.0.1\',\n             \'port\': 5433,\n             \'user\': \'some_user\',\n             \'password\': \'some_password\',\n             \'database\': \'vdb\',\n             \'connection_load_balance\': True}\n\n# Server enables load balancing\nasync with aiovertica.connect(**conn_info) as conn:\n    cur = conn.cursor()\n    await cur.execute("SELECT NODE_NAME FROM V_MONITOR.CURRENT_SESSION")\n    result = await cur.fetchone()\n    print("Client connects to primary node:", result[0])\n    await cur.execute("SELECT SET_LOAD_BALANCE_POLICY(\'ROUNDROBIN\')")\n\nasync with aiovertica.connect(**conn_info) as conn:\n    cur = conn.cursor()\n    await cur.execute("SELECT NODE_NAME FROM V_MONITOR.CURRENT_SESSION")\n    result = await cur.fetchone()\n    print("Client redirects to node:", result[0])\n\n## Output\n#  Client connects to primary node: v_vdb_node0003\n#  Client redirects to node: v_vdb_node0005\n```\n\nAnother way to set connection properties is passing a connection string to the keyword parameter `dsn` of `vertica_python.connect(dsn=\'...\', **kwargs)`. The connection string is of the form:\n```\nvertica://(user):(password)@(host):(port)/(database)?(arg1=val1&arg2=val2&...)\n```\nThe connection string would be parsed by `vertica_python.parse_dsn(connection_str)`, and the parsing result (a dictionary of keywords and values) would be merged with _kwargs_. If the same keyword is specified in both the sources, the _kwargs_ value overrides the parsed _dsn_ value. The `(arg1=val1&arg2=val2&...)` section can handle string/numeric/boolean values, blank and invalid value would be ignored.\n\n```python\nimport aiovertica\n\nconnection_str = (\n    \'vertica://admin@localhost:5433/db1?connection_load_balance=True&connection_timeout=1.5&session_label=vpclient+123%7E456\'\n)\nprint(aiovertica.parse_dsn(connection_str))\n# {\'user\': \'admin\', \'host\': \'localhost\', \'port\': 5433, \'database\': \'db1\',\n#  \'connection_load_balance\': True, \'connection_timeout\': 1.5, \'session_label\': \'vpclient 123~456\'}\n\nadditional_info = {\n    \'password\': \'some_password\', \n    \'backup_server_node\': [\'10.6.7.123\', (\'10.20.82.77\', 6000)]  # invalid value to be set in a connection string\n}\n\nasync with aiovertica.connect(dsn=connection_str, **additional_info) as conn:\n   # do things\n```\n\n\n**Stream query results**:\n\n```python\ncur = connection.cursor()\nawait cur.execute("SELECT * FROM a_table LIMIT 2")\n\nasync for row in cur.iterate():\n    print(row)\n# [ 1, \'some text\', datetime.datetime(2014, 5, 18, 6, 47, 1, 928014) ]\n# [ 2, \'something else\', None ]\n\n```\nStreaming is recommended if you want to further process each row, save the results in a non-list/dict format (e.g. Pandas DataFrame), or save the results in a file.\n\n\n**In-memory results as list**:\n\n```python\ncur = connection.cursor()\nawait cur.execute("SELECT * FROM a_table LIMIT 2")\nawait cur.fetchall()\n# [ [1, \'something\'], [2, \'something_else\'] ]\n```\n\n\n**In-memory results as dictionary**:\n\n```python\ncur = connection.cursor(\'dict\')\nawait cur.execute("SELECT * FROM a_table LIMIT 2")\nawait cur.fetchall()\n# [ {\'id\': 1, \'value\': \'something\'}, {\'id\': 2, \'value\': \'something_else\'} ]\nawait connection.close()\n```\n\n\n**Query using named parameters or format parameters**:\n\nvertica-python can automatically convert Python objects to SQL literals: using this feature your code will be more robust and reliable to prevent SQL injection attacks.\n\nPrerequisites: Only SQL literals (i.e. query values) should be bound via these methods: they shouldn’t be used to merge table or field names to the query (_vertica-python_ will try quoting the table name as a string value, generating invalid SQL as it is actually a SQL Identifier). If you need to generate dynamically SQL queries (for instance choosing dynamically a table name) you have to construct the full query yourself.\n\nVariables can be specified with named (__:name__) placeholders.\n```python\ncur = connection.cursor()\ndata = {\'propA\': 1, \'propB\': \'stringValue\'}\nawait cur.execute("SELECT * FROM a_table WHERE a = :propA AND b = :propB", data)\n# converted into a SQL command similar to: "SELECT * FROM a_table WHERE a = 1 AND b = \'stringValue\'"\n\nawait cur.fetchall()\n# [ [1, \'stringValue\'] ]\n```\n\nVariables can also be specified with positional format (__%s__) placeholders. The placeholder __must always be a %s__, even if a different placeholder (such as a %d for integers or %f for floats) may look more appropriate. __Never__ use Python string concatenation (+) or string parameters interpolation (%) to pass variables to a SQL query string.\n```python\ncur = connection.cursor()\ndata = (1, "O\'Reilly")\nawait cur.execute("SELECT * FROM a_table WHERE a = %s AND b = %s" % data) # WRONG: % operator\nawait cur.execute("SELECT * FROM a_table WHERE a = %d AND b = %s", data)  # WRONG: %d placeholder\nawait cur.execute("SELECT * FROM a_table WHERE a = %s AND b = %s", data)  # correct\n# converted into a SQL command similar to: "SELECT * FROM a_table WHERE a = 1 AND b = \'O\'\'Reilly\'"\n\nawait cur.fetchall()\n# [ [1, "O\'Reilly"] ]\n```\n\nThe placeholder must not be quoted. _vertica-python_ will add quotes where needed.\n```python\ncur = connection.cursor()\nawait cur.execute("INSERT INTO table VALUES (\':propA\')", {\'propA\': "someString"}) # WRONG\nawait cur.execute("INSERT INTO table VALUES (:propA)", {\'propA\': "someString"})   # correct\nawait cur.execute("INSERT INTO table VALUES (\'%s\')", ("someString",)) # WRONG\nawait cur.execute("INSERT INTO table VALUES (%s)", ("someString",))   # correct\n```\n\n_vertica-python_ supports default mapping for many standard Python types. It is possible to adapt new Python types to SQL literals via `Cursor.register_sql_literal_adapter(py_class_or_type, adapter_function)` function. Example:\n```python\nclass Point(object):\n    def __init__(self, x, y):\n        self.x = x\n        self.y = y\n\n# Adapter should return a string value\ndef adapt_point(point):\n    return "STV_GeometryPoint({},{})".format(point.x, point.y)\n\ncur = conn.cursor()\ncur.register_sql_literal_adapter(Point, adapt_point)\n\nawait cur.execute("INSERT INTO geom_data (geom) VALUES (%s)", [Point(1.23, 4.56)])\nawait cur.execute("select ST_asText(geom) from geom_data")\nawait cur.fetchall()\n# [[\'POINT (1.23 4.56)\']]\n```\n\nTo help you debug the binding process during Cursor.execute*(), `Cursor.object_to_sql_literal(py_object)` function can be used to inspect the SQL literal string converted from a Python object.\n```python\ncur = conn.cursor\ncur.object_to_sql_literal("O\'Reilly")  # "\'O\'\'Reilly\'"\ncur.object_to_sql_literal(None)  # "NULL"\ncur.object_to_sql_literal(True)  # "True"\ncur.object_to_sql_literal(Decimal("10.00000"))  # "10.00000"\ncur.object_to_sql_literal(datetime.date(2018, 9, 7))  # "\'2018-09-07\'"\ncur.object_to_sql_literal(Point(-71.13, 42.36))  # "STV_GeometryPoint(-71.13,42.36)" if you registered in previous step\n```\n\n\n**Query using server-side prepared statements**:\n\nVertica server-side prepared statements let you define a statement once and then run it many times with different parameters. Placeholders in the statement are represented by question marks (?). Server-side prepared statements are useful for preventing SQL injection attacks.\n\n```python\nimport aiovertica\n\n# Enable using server-side prepared statements at connection level\nconn_info = {\n    \'host\': \'127.0.0.1\',\n    \'user\': \'some_user\',\n    \'password\': \'some_password\',\n    \'database\': \'a_database\',\n    \'use_prepared_statements\': True,\n}\n\nasync with aiovertica.connect(**conn_info) as connection:\n    cur = connection.cursor()\n    await cur.execute("CREATE TABLE tbl (a INT, b VARCHAR)")\n    await cur.execute("INSERT INTO tbl VALUES (?, ?)", [1, \'aa\'])\n    await cur.execute("INSERT INTO tbl VALUES (?, ?)", [2, \'bb\'])\n    await cur.executemany("INSERT INTO tbl VALUES (?, ?)", [(3, \'foo\'), (4, \'xx\'), (5, \'bar\')])\n    await cur.execute("COMMIT")\n\n    await cur.execute("SELECT * FROM tbl WHERE a>=? AND a<=? ORDER BY a", (2,4))\n    await cur.fetchall()\n    # [[2, \'bb\'], [3, \'foo\'], [4, \'xx\']]\n```\n\nVertica does not support executing a command string containing multiple statements using server-side prepared statements. You can set ```use_prepared_statements``` option in ```cursor.execute*()``` functions to override the connection level setting.\n\n```python\nimport aiovertica\n\n# Enable using server-side prepared statements at connection level\nconn_info = {\n    \'host\': \'127.0.0.1\',\n    \'user\': \'some_user\',\n    \'password\': \'some_password\',\n    \'database\': \'a_database\',\n    \'use_prepared_statements\': True,\n}\n\nasync with aiovertica.connect(**conn_info) as connection:\n    cur = connection.cursor()\n    await cur.execute("CREATE TABLE tbl (a INT, b VARCHAR)")\n\n    # Executing compound statements\n    await cur.execute("INSERT INTO tbl VALUES (?, ?); COMMIT", [1, \'aa\'])\n    # Error message: Cannot insert multiple commands into a prepared statement\n\n    # Disable prepared statements but forget to change placeholders (?)\n    await cur.execute("INSERT INTO tbl VALUES (?, ?); COMMIT;", [1, \'aa\'], use_prepared_statements=False)\n    # TypeError: not all arguments converted during string formatting\n\n    await cur.execute("INSERT INTO tbl VALUES (%s, %s); COMMIT;", [1, \'aa\'], use_prepared_statements=False)\n    await cur.execute("INSERT INTO tbl VALUES (:a, :b); COMMIT;", {\'a\': 2, \'b\': \'bb\'}, use_prepared_statements=False)\n\n# Disable using server-side prepared statements at connection level\nconn_info[\'use_prepared_statements\'] = False\nasync with aiovertica.connect(**conn_info) as connection:\n    cur = connection.cursor()\n\n    # Try using prepared statements\n    await cur.execute("INSERT INTO tbl VALUES (?, ?)", [3, \'foo\'])\n    # TypeError: not all arguments converted during string formatting\n\n    await cur.execute("INSERT INTO tbl VALUES (?, ?)", [3, \'foo\'], use_prepared_statements=True)\n\n    # Query using named parameters\n    await cur.execute("SELECT * FROM tbl WHERE a>=:n1 AND a<=:n2 ORDER BY a", {\'n1\': 2, \'n2\': 4})\n    await cur.fetchall()\n    # [[2, \'bb\'], [3, \'foo\']]\n```\nNote: In other drivers, the batch insert is converted into a COPY statement by using prepared statements. vertica-python currently does not support that.\n\n**Insert and commits**:\n\n```python\ncur = connection.cursor()\n\n# inline commit (when \'use_prepared_statements\' is False)\nawait cur.execute("INSERT INTO a_table (a, b) VALUES (1, \'aa\'); commit;")\n\n# commit in execution\nawait cur.execute("INSERT INTO a_table (a, b) VALUES (1, \'aa\')")\nawait cur.execute("INSERT INTO a_table (a, b) VALUES (2, \'bb\')")\nawait cur.execute("commit;")\n\n# connection.commit()\nawait cur.execute("INSERT INTO a_table (a, b) VALUES (1, \'aa\')")\nawait connection.commit()\n\n# connection.rollback()\nawait cur.execute("INSERT INTO a_table (a, b) VALUES (0, \'bad\')")\nawait connection.rollback()\n```\n\n**Autocommit**:\n\nSession parameter AUTOCOMMIT can be configured by the connection option and the `Connection.autocommit` read/write attribute:\n```python\nimport aiovertica\n\n# Enable autocommit at startup\nconn_info = {\n    \'host\': \'127.0.0.1\',\n    \'user\': \'some_user\',\n    \'password\': \'some_password\',\n    \'database\': \'a_database\',\n    # autocommit is off by default\n    \'autocommit\': True,\n}\n\nasync with aiovertica.connect(**conn_info) as connection:\n    # Check current session autocommit setting\n    print(connection.autocommit)    # should be True\n    # If autocommit is True, statements automatically commit their transactions when they complete.\n    \n    # Set autocommit setting with attribute\n    connection.autocommit = False\n    print(connection.autocommit)    # should be False\n    # If autocommit is False, the methods commit() or rollback() must be manually invoked to terminate the transaction.\n```\n\n> #### WARNING: set attr `autocommit` with instance may not work correctly in the current version\n\nTo set AUTOCOMMIT to a new value, vertica-python uses `Cursor.execute()` to execute a command internally, and that would clear your previous query results, so be sure to call `Cursor.fetch*()` to save your results before you set autocommit.\n\n\n**Copy**:\n\nThere are 2 methods to do copy:\n\nMethod 1: "COPY FROM STDIN" sql with Cursor.copy()\n```python\ncur = connection.cursor()\nawait cur.copy("COPY test_copy (id, name) from stdin DELIMITER \',\' ",  csv)\n```\n\nWhere `csv` is either a string or a file-like object (specifically, any object with a `read()` method). If using a file, the data is streamed (in chunks of `buffer_size` bytes, which defaults to 128 * 2 ** 10).\n\n```python\nwith open("/tmp/binary_file.csv", "rb") as fs:\n    await cursor.copy(\n        "COPY table(field1, field2) FROM STDIN DELIMITER \',\' ENCLOSED BY \'\\"\'",\n        fs,\n        buffer_size=65536\n    )\n```\n\nMethod 2: "COPY FROM LOCAL" sql with Cursor.execute() \n\n```python\nimport sys\nimport aiovertica\n\nconn_info = {\'host\': \'127.0.0.1\',\n             \'user\': \'some_user\',\n             \'password\': \'some_password\',\n             \'database\': \'a_database\',\n             # False by default\n             #\'disable_copy_local\': True,\n             # Don\'t support executing COPY LOCAL operations with prepared statements\n             \'use_prepared_statements\': False\n             }\n\nasync with aiovertica.connect(**conn_info) as connection:\n    cur = connection.cursor()\n    \n    # Copy from local file\n    await cur.execute(\n        "COPY table(field1, field2) FROM LOCAL"\n        " \'data_Jan_*.csv\',\'data_Feb_01.csv\' DELIMITER \',\'"\n        " REJECTED DATA \'path/to/write/rejects.txt\'"\n        " EXCEPTIONS \'path/to/write/exceptions.txt\'",\n        buffer_size=65536\n    )\n    result = await cur.fetchall()\n    print("Rows loaded:", result)\n    \n    # Copy from local stdin\n    await cur.execute(\n        "COPY table(field1, field2) FROM LOCAL STDIN DELIMITER \',\'",\n        copy_stdin=sys.stdin\n    )\n    result = await cur.fetchall()\n    print("Rows loaded:", result)\n\n    # Copy from local stdin (compound statements)\n    with open(\'f1.csv\', \'r\') as fs1, open(\'f2.csv\', \'r\') as fs2:\n        await cur.execute(\n            "COPY tlb1(field1, field2) FROM LOCAL STDIN DELIMITER \',\';"\n            "COPY tlb2(field1, field2) FROM LOCAL STDIN DELIMITER \',\';",\n            copy_stdin=[fs1, fs2], buffer_size=65536\n        )\n    result = await cur.fetchall()\n    print("Rows loaded 1:", result)\n    await cur.nextset()\n    result = await cur.fetchall()\n    print("Rows loaded 2:", result)\n```\nWhen connection option `disable_copy_local` set to True, disables COPY LOCAL operations, including copying data from local files/stdin and using local files to store data and exceptions. You can use this property to prevent users from writing to and copying from files on a Vertica host, including an MC host. Note that this property doesn\'t apply to `Cursor.copy()`.\n\nThe data for copying from/writing to local files is streamed in chunks of `buffer_size` bytes, which defaults to 128 * 2 ** 10.\n\nWhen executing "COPY FROM LOCAL STDIN", `copy_stdin` should be a file-like object or a list of file-like objects (specifically, any object with a `read()` method).\n\n**Cancel the current database operation**:\n\n`Connection.cancel()` interrupts the processing of the current operation. Interrupting query execution will cause the cancelled method to raise a `vertica_python.errors.QueryCanceled`. If no query is being executed, it does nothing. You can call this function from a different thread/process than the one currently executing a database operation.\n\n```python\nfrom multiprocessing import Process\nimport asyncio\nimport aiovertica\n\nasync def cancel_query(connection, timeout=5):\n    await asyncio.sleep(timeout)\n    await connection.cancel()\n\n# Example 1: Cancel the query before Cursor.execute() return.\n#            The query stops executing in a shorter time after the cancel message is sent.\nasync with aiovertica.connect(**conn_info) as conn:\n    cur = conn.cursor()\n\n    # Call cancel() from a different process\n    p1 = Process(target=cancel_query, args=(conn,))\n    p1.start()\n\n    try:\n        cur.execute("<Long running query>")\n    except aiovertica.errors.QueryCanceled as e:\n        pass\n\n    p1.join()\n    \n# Example 2: Cancel the query after Cursor.execute() return.\n#            Less number of rows read after the cancel message is sent.\nwith aiovertica.connect(**conn_info) as conn:\n    cur = conn.cursor()\n    await cur.execute("SELECT id, time FROM large_table")\n    nCount = 0\n    try:\n        result = await cur.fetchone()\n        while result:\n            nCount += 1\n            if nCount == 100:\n                conn.cancel()\n            result = await cur.fetchone()\n    except aiovertica.errors.QueryCanceled as e:\n        pass\n        # nCount is less than the number of rows in large_table\n\n```\n\n\n\n## Rowcount oddities\n\nvertica_python behaves a bit differently than dbapi when returning rowcounts.\n\nAfter a select execution, the rowcount will be -1, indicating that the row count is unknown. The rowcount value will be updated as data is streamed.\n\n```python\nawait cur.execute(\'SELECT 10 things\')\n\ncur.rowcount == -1  # indicates unknown rowcount\n\nawait cur.fetchone()\ncur.rowcount == 1\nawait cur.fetchone()\ncur.rowcount == 2\nawait cur.fetchall()\ncur.rowcount == 10\n```\n\nAfter an insert/update/delete, the rowcount will be returned as a single element row:\n\n```python\nawait cur.execute("DELETE 3 things")\n\ncur.rowcount == -1  # indicates unknown rowcount\nresult = await cur.fetchone()\nresult[0] == 3\n```\n\n## Nextset\n\nIf you execute multiple statements in a single call to execute(), you can use cursor.nextset() to retrieve all of the data.\n\n```python\nawait cur.execute(\'SELECT 1; SELECT 2;\')\n\nawait cur.fetchone()\n# [1]\nawait cur.fetchone()\n# None\n\nawait cur.nextset()\n# True\n\nawait cur.fetchone()\n# [2]\nawait cur.fetchone()\n# None\n\nawait cur.nextset()\n# None\n```\n\n## UTF-8 encoding issues\n\nWhile Vertica expects varchars stored to be UTF-8 encoded, sometimes invalid strings get into the database. You can specify how to handle reading these characters using the unicode_error connection option. This uses the same values as the unicode type (https://docs.python.org/2/library/functions.html#unicode)\n\n```python\nimport aiovertica\n\nconn = await aiovertica.connect({..., \'unicode_error\': \'strict\'})\ncur = conn.cursor()\nawait cur.execute(r"SELECT E\'\\xC2\'")\nawait cur.fetchone()\n# caught \'utf8\' codec can\'t decode byte 0xc2 in position 0: unexpected end of data\n\nconn = await aiovertica.connect({..., \'unicode_error\': \'strict\'})\ncur = conn.cursor()\nawait cur.execute(r"SELECT E\'\\xC2\'")\nawait cur.fetchone()\n# �\n\nconn = await aiovertica.connect({..., \'unicode_error\': \'strict\'})\ncur = conn.cursor()\nawait cur.execute(r"SELECT E\'\\xC2\'")\nawait cur.fetchone()\n# \n```\n\n## License\n\nApache 2.0 License, please see `LICENSE` for details.\n\n## Contributing guidelines\n\nHave a bug or an idea? Please see `CONTRIBUTING.md` for details.\n\n## Acknowledgements\n\nI would like to express my deep gratitude to the team \nthat worked on the [vertica_python](https://github.com/vertica/vertica-python/) for writing at least a synchronous driver - this greatly \nsimplified the work. I have tried to keep all the licenses and the list of authors from \nthe original project.\n',
    'author': 'Ivan Galin',
    'author_email': 'i.galin@devartsteam.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/i8enn/aiovertica',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)
